from .. import ssa_types
from ..constraints import IntConstraint, return_

from .base import BaseOp

class FAdd(BaseOp):
    def __init__(self, parent, args):
        BaseOp.__init__(self, parent, args)
        self.rval = parent.makeVariable(args[0].type, origin=self)
class FDiv(BaseOp):
    def __init__(self, parent, args):
        BaseOp.__init__(self, parent, args)
        self.rval = parent.makeVariable(args[0].type, origin=self)
class FMul(BaseOp):
    def __init__(self, parent, args):
        BaseOp.__init__(self, parent, args)
        self.rval = parent.makeVariable(args[0].type, origin=self)
class FRem(BaseOp):
    def __init__(self, parent, args):
        BaseOp.__init__(self, parent, args)
        self.rval = parent.makeVariable(args[0].type, origin=self)
class FSub(BaseOp):
    def __init__(self, parent, args):
        BaseOp.__init__(self, parent, args)
        self.rval = parent.makeVariable(args[0].type, origin=self)

# Unary, unlike the others
class FNeg(BaseOp):
    def __init__(self, parent, args):
        BaseOp.__init__(self, parent, args)
        self.rval = parent.makeVariable(args[0].type, origin=self)

class FCmp(BaseOp):
    def __init__(self, parent, args, NaN_val):
        BaseOp.__init__(self, parent, args)
        self.rval = parent.makeVariable(ssa_types.SSA_INT, origin=self)
        self.NaN_val = NaN_val

    def propagateConstraints(self, x, y):
        return return_(IntConstraint.range(32, -1, 1))



#!/usr/bin/env python2
from __future__ import print_function

import os.path
import time, random, subprocess, functools

import Krakatau
import Krakatau.ssa
from Krakatau.error import ClassLoaderError
from Krakatau.environment import Environment
from Krakatau.java import javaclass, visitor
from Krakatau.java.stringescape import escapeString
from Krakatau.verifier.inference_verifier import verifyBytecode
from Krakatau import script_util

def findJRE():
    try:
        home = os.environ.get('JAVA_HOME')
        if home:
            path = os.path.join(home, 'jre', 'lib', 'rt.jar')
            if os.path.isfile(path):
                return path

            # For macs
            path = os.path.join(home, 'bundle', 'Classes', 'classes.jar')
            if os.path.isfile(path):
                return path

        # Ubuntu and co
        out = subprocess.Popen(['update-java-alternatives', '-l'], stdout=subprocess.PIPE).communicate()[0]
        basedir = out.split('\n')[0].rpartition(' ')[-1]
        path = os.path.join(basedir, 'jre', 'lib', 'rt.jar')
        if os.path.isfile(path):
            return path
    except Exception as e:
        pass

def _stats(s):
    bc = len(s.blocks)
    vc = sum(len(b.unaryConstraints) for b in s.blocks)
    return '{} blocks, {} variables'.format(bc,vc)

def _print(s):
    from Krakatau.ssa.printer import SSAPrinter
    return SSAPrinter(s).print_()

def makeGraph(opts, m):
    v = verifyBytecode(m.code)
    s = Krakatau.ssa.ssaFromVerified(m.code, v, opts)

    if s.procs:
        # s.mergeSingleSuccessorBlocks()
        # s.removeUnusedVariables()
        s.inlineSubprocs()

    # print(_stats(s))
    s.condenseBlocks()
    s.mergeSingleSuccessorBlocks()
    s.removeUnusedVariables()
    # print(_stats(s))

    s.copyPropagation()
    s.abstractInterpert()
    s.disconnectConstantVariables()

    s.simplifyThrows()
    s.simplifyCatchIgnored()
    s.mergeSingleSuccessorBlocks()
    s.mergeSingleSuccessorBlocks()
    s.removeUnusedVariables()
    # print(_stats(s))
    return s

def deleteUnusued(cls):
    # Delete attributes we aren't going to use
    # pretty hackish, but it does help when decompiling large jars
    for e in cls.fields + cls.methods:
        del e.class_, e.attributes, e.static
    for m in cls.methods:
        del m.native, m.abstract, m.isConstructor
        del m.code
    del cls.version, cls.this, cls.super, cls.env
    del cls.interfaces_raw, cls.cpool
    del cls.attributes

def decompileClass(path=[], targets=None, outpath=None, skip_errors=False, add_throws=False, magic_throw=False):
    out = script_util.makeWriter(outpath, '.java')

    e = Environment()
    for part in path:
        e.addToPath(part)

    start_time = time.time()
    # random.shuffle(targets)
    with e, out:
        printer = visitor.DefaultVisitor()
        for i,target in enumerate(targets):
            print('processing target {}, {} remaining'.format(target, len(targets)-i))

            try:
                c = e.getClass(target.decode('utf8'))
                makeGraphCB = functools.partial(makeGraph, magic_throw)
                source = printer.visit(javaclass.generateAST(c, makeGraphCB, skip_errors, add_throws=add_throws))
            except Exception as err:
                if not skip_errors:
                    raise
                if isinstance(err, ClassLoaderError):
                    print('Failed to decompile {} due to missing or invalid class {}'.format(target, err.data))
                else:
                    import traceback
                    print(traceback.format_exc())
                continue

            # The single class decompiler doesn't add package declaration currently so we add it here
            if '/' in target:
                package = 'package {};\n\n'.format(escapeString(target.replace('/','.').rpartition('.')[0]))
                source = package + source

            filename = out.write(c.name.encode('utf8'), source)
            print('Class written to', filename)
            print(time.time() - start_time, ' seconds elapsed')
            deleteUnusued(c)
        print(len(e.classes) - len(targets), 'extra classes loaded')

if __name__== "__main__":
    print(script_util.copyright)

    import argparse
    parser = argparse.ArgumentParser(description='Krakatau decompiler and bytecode analysis tool')
    parser.add_argument('-path',action='append',help='Semicolon seperated paths or jars to search when loading classes')
    parser.add_argument('-out',help='Path to generate source files in')
    parser.add_argument('-nauto', action='store_true', help="Don't attempt to automatically locate the Java standard library. If enabled, you must specify the path explicitly.")
    parser.add_argument('-r', action='store_true', help="Process all files in the directory target and subdirectories")
    parser.add_argument('-skip', action='store_true', help="Upon errors, skip class or method and continue decompiling")
    parser.add_argument('-xmagicthrow', action='store_true')
    parser.add_argument('target',help='Name of class or jar file to decompile')
    args = parser.parse_args()

    path = []
    if not args.nauto:
        print('Attempting to automatically locate the standard library...')
        found = findJRE()
        if found:
            print('Found at ', found)
            path.append(found)
        else:
            print('Unable to find the standard library')

    if args.path:
        for part in args.path:
            path.extend(part.split(';'))

    if args.target.endswith('.jar'):
        path.append(args.target)

    targets = script_util.findFiles(args.target, args.r, '.class')
    targets = map(script_util.normalizeClassname, targets)
    decompileClass(path, targets, args.out, args.skip, magic_throw=args.xmagicthrow)



#!/usr/bin/env python2
from __future__ import print_function

import functools
import os.path
import time, zipfile, sys

try:
    from StringIO import StringIO
except ImportError:
    from io import StringIO


import Krakatau
from Krakatau import script_util
from Krakatau.classfileformat.reader import Reader
from Krakatau.classfileformat.classdata import ClassData
from Krakatau.assembler.disassembly import Disassembler

def readArchive(archive, name):
    with archive.open(name.decode('utf8')) as f:
        return f.read()

def readFile(filename):
    with open(filename, 'rb') as f:
        return f.read()

def disassembleSub(readTarget, out, targets, roundtrip=False, outputClassName=True):
    start_time = time.time()
    with out:
        for i, target in enumerate(targets):
            print('processing target {}, {}/{} remaining'.format(target, len(targets)-i, len(targets)))

            data = readTarget(target)
            clsdata = ClassData(Reader(data))

            if outputClassName:
                name = clsdata.pool.getclsutf(clsdata.this)
            else:
                name = target.rpartition('.')[0] or target

            output = StringIO()
            # output = sys.stdout
            Disassembler(clsdata, output.write, roundtrip=roundtrip).disassemble()

            filename = out.write(name, output.getvalue())
            if filename is not None:
                print('Class written to', filename)
                print(time.time() - start_time, ' seconds elapsed')

if __name__== "__main__":
    print(script_util.copyright)

    import argparse
    parser = argparse.ArgumentParser(description='Krakatau decompiler and bytecode analysis tool')
    parser.add_argument('-out', help='Path to generate files in')
    parser.add_argument('-r', action='store_true', help="Process all files in the directory target and subdirectories")
    parser.add_argument('-path', help='Jar to look for class in')
    parser.add_argument('-roundtrip', action='store_true', help='Create assembly file that can roundtrip to original binary.')
    parser.add_argument('target', help='Name of class or jar file to decompile')
    args = parser.parse_args()

    targets = script_util.findFiles(args.target, args.r, '.class')

    jar = args.path
    if jar is None and args.target.endswith('.jar'):
        jar = args.target

    out = script_util.makeWriter(args.out, '.j')
    if jar is not None:
        with zipfile.ZipFile(jar, 'r') as archive:
            readFunc = functools.partial(readArchive, archive)
            disassembleSub(readFunc, out, targets, roundtrip=args.roundtrip)
    else:
        disassembleSub(readFile, out, targets, roundtrip=args.roundtrip, outputClassName=False)



# Override this to rename classes
class DefaultVisitor(object):
    def visit(self, obj):
        return obj.print_(self, self.visit)

    # Experimental - don't use!
    def toTree(self, obj):
        if obj is None:
            return None
        return obj.tree(self, self.toTree)

    def className(self, name): return name
    def methodName(self, cls, name, desc): return name
    def fieldName(self, cls, name, desc): return name



import collections

from .. import graph_util
from ..ssa import objtypes
from ..ssa.objtypes import BExpr, BoolTT, ByteTT, CharTT, IntTT, ShortTT

from . import ast

# Class union-find data structure except that we don't bother with weighting trees and singletons are implicit
# Also, booleans are forced to be seperate roots
FORCED_ROOTS = True, False
class UnionFind(object):
    def __init__(self):
        self.d = {}

    def find(self, x):
        if x not in self.d:
            return x
        path = [x]
        while path[-1] in self.d:
            path.append(self.d[path[-1]])
        root = path.pop()
        for y in path:
            self.d[y] = root
        return root

    def union(self, x, x2):
        if x is None or x2 is None:
            return
        root1, root2 = self.find(x), self.find(x2)
        if root2 in FORCED_ROOTS:
            root1, root2 = root2, root1
        if root1 != root2 and root2 not in FORCED_ROOTS:
            self.d[root2] = root1

##############################################################
def visitStatementTree(scope, callback, catchcb=None):
    for item in scope.statements:
        for sub in item.getScopes():
            visitStatementTree(sub, callback, catchcb)
        if item.expr is not None:
            callback(item, item.expr)
        if catchcb is not None and isinstance(item, ast.TryStatement):
            for pair in item.pairs:
                catchcb(pair[0])

int_tags = frozenset(map(objtypes.baset, [IntTT, ShortTT, CharTT, ByteTT, BoolTT]))
array_tags = frozenset(map(objtypes.baset, [ByteTT, BoolTT]) + [objtypes.BExpr])

# Fix int/bool and byte[]/bool[] vars
def boolizeVars(root, arg_vars):
    varlist = []
    sets = UnionFind()

    def visitExpr(expr, forceExact=False):
        # see if we have to merge
        if isinstance(expr, ast.Assignment) or isinstance(expr, ast.BinaryInfix) and expr.opstr in ('==','!=','&','|','^'):
            subs = [visitExpr(param) for param in expr.params]
            sets.union(*subs) # these operators can work on either type but need the same type on each side
        elif isinstance(expr, ast.ArrayAccess):
            sets.union(False, visitExpr(expr.params[1])) # array index is int only
        elif isinstance(expr, ast.BinaryInfix) and expr.opstr in ('* / % + - << >> >>>'):
            sets.union(False, visitExpr(expr.params[0])) # these operators are int only
            sets.union(False, visitExpr(expr.params[1]))

        if isinstance(expr, ast.Local):
            tag, dim = objtypes.baset(expr.dtype), objtypes.dim(expr.dtype)
            if (dim == 0 and tag in int_tags) or (dim > 0 and tag in array_tags):
                # the only "unknown" vars are bexpr[] and ints. All else have fixed types
                if forceExact or (tag != BExpr and tag != objtypes.baset(IntTT)):
                    sets.union(tag == objtypes.baset(BoolTT), expr)
                varlist.append(expr)
                return sets.find(expr)
        elif isinstance(expr, ast.Literal):
            if expr.dtype == IntTT and expr.val not in (0,1):
                return False
            return None # if val is 0 or 1, or the literal is a null, it is freely convertable
        elif isinstance(expr, ast.Assignment) or (isinstance(expr, ast.BinaryInfix) and expr.opstr in ('&','|','^')):
            return subs[0]
        elif isinstance(expr, (ast.ArrayAccess, ast.Parenthesis, ast.UnaryPrefix)):
            return visitExpr(expr.params[0])
        elif expr.dtype is not None and objtypes.baset(expr.dtype) != BExpr:
            return expr.dtype[0] == objtypes.baset(BoolTT)
        return None

    def visitStatement(item, expr):
        root = visitExpr(expr)
        if isinstance(item, ast.ReturnStatement):
            forced_val = (objtypes.baset(item.tt) == objtypes.baset(BoolTT))
            sets.union(forced_val, root)
        elif isinstance(item, ast.SwitchStatement):
            sets.union(False, root) # Switch must take an int, not a bool

    for expr in arg_vars:
        visitExpr(expr, forceExact=True)
    visitStatementTree(root, callback=visitStatement)

    # Fix the propagated types
    for var in set(varlist):
        tag, dim = objtypes.baset(var.dtype), objtypes.dim(var.dtype)
        assert tag in int_tags or (dim>0 and tag == BExpr)
        # make everything bool which is not forced to int
        if sets.find(var) != False:
            var.dtype = objtypes.withDimInc(BoolTT, dim)
        elif dim > 0:
            var.dtype = objtypes.withDimInc(ByteTT, dim)

    # Fix everything else back up
    def fixExpr(item, expr):
        for param in expr.params:
            fixExpr(None, param)

        if isinstance(expr, ast.Assignment):
            left, right = expr.params
            if objtypes.baset(left.dtype) in int_tags and objtypes.dim(left.dtype) == 0:
                if not ast.isPrimativeAssignable(right.dtype, left.dtype):
                    expr.params = [left, ast.makeCastExpr(left.dtype, right)]
        elif isinstance(expr, ast.BinaryInfix):
            a, b = expr.params
            # shouldn't need to do anything here for arrays
            if expr.opstr in '== != & | ^' and a.dtype == BoolTT or b.dtype == BoolTT:
                expr.params = [ast.makeCastExpr(BoolTT, v) for v in expr.params]
    visitStatementTree(root, callback=fixExpr)

# Fix vars of interface/object type
# TODO: do this properly
def interfaceVars(env, root, arg_vars):
    varlist = []
    consts = {}
    assigns = collections.defaultdict(list)

    def isInterfaceVar(expr):
        if not isinstance(expr, ast.Local) or not objtypes.isBaseTClass(expr.dtype):
            return False
        if objtypes.className(expr.dtype) == objtypes.className(objtypes.ObjectTT):
            return True
        return env.isInterface(objtypes.className(expr.dtype))

    def updateConst(var, tt):
        varlist.append(var)
        if var not in consts:
            consts[var] = tt
        else:
            consts[var] = objtypes.commonSupertype(env, [consts[var], tt])

    def visitStatement(item, expr):
        if isinstance(expr, ast.Assignment) and objtypes.isBaseTClass(expr.dtype):
            left, right = expr.params
            if isInterfaceVar(left):
                if isInterfaceVar(right):
                    assigns[left].append(right)
                    varlist.append(right)
                    varlist.append(left)
                else:
                    updateConst(left, right.dtype)

    def visitCatchDecl(decl):
        updateConst(decl.local, decl.typename.dtype)

    for expr in arg_vars:
        if objtypes.isBaseTClass(expr.dtype):
            updateConst(expr, expr.dtype)
    visitStatementTree(root, callback=visitStatement, catchcb=visitCatchDecl)

    # Now calculate actual types and fix
    newtypes = {}

    # visit variables in topological order. Doesn't handle case of loops, but this is a temporary hack anyway
    order = graph_util.topologicalSort(varlist, lambda v:assigns[v])
    for var in order:
        assert var not in newtypes

        tts = [newtypes.get(right, objtypes.ObjectTT) for right in assigns[var]]
        if var in consts:
            tts.append(consts[var])
        newtypes[var] = newtype = objtypes.commonSupertype(env, tts)
        if newtype != objtypes.ObjectTT and newtype != var.dtype and newtype != objtypes.NullTT:
            # assert objtypes.baset(var.dtype) == objtypes.baset(objtypes.ObjectTT)
            var.dtype = newtype



from ...verifier.descriptors import parseMethodDescriptor

from .. import constraints, excepttypes, objtypes
from ..constraints import ObjectConstraint, returnOrThrow, throw
from ..ssa_types import SSA_OBJECT, verifierToSSAType

from .base import BaseOp

class Invoke(BaseOp):
    has_side_effects = True

    def __init__(self, parent, instr, info, args, isThisCtor, target_tt):
        super(Invoke, self).__init__(parent, args, makeException=True)

        self.instruction = instr
        self.target, self.name, self.desc = info
        self.isThisCtor = isThisCtor # whether this is a ctor call for the current class
        self.target_tt = target_tt
        vtypes = parseMethodDescriptor(self.desc)[1]

        dtype = None
        if vtypes:
            stype = verifierToSSAType(vtypes[0])
            dtype = objtypes.verifierToSynthetic(vtypes[0])
            cat = len(vtypes)
            # clone() on an array type is known to always return that type, rather than any Object
            if self.name == "clone" and target_tt[1] > 0:
                dtype = target_tt

            self.rval = parent.makeVariable(stype, origin=self)
            self.returned = [self.rval] + [None]*(cat-1)
        else:
            self.rval, self.returned = None, []

        # just use a fixed constraint until we can do interprocedural analysis
        # output order is rval, exception, defined by BaseOp.getOutputs
        env = parent.env
        self.eout = ObjectConstraint.fromTops(env, [objtypes.ThrowableTT], [], nonnull=True)
        self.eout_npe = ObjectConstraint.fromTops(env, [excepttypes.NullPtr], [], nonnull=True)
        if self.rval is not None:
            if self.rval.type == SSA_OBJECT:
                supers, exact = objtypes.declTypeToActual(env, dtype)
                self.rout = ObjectConstraint.fromTops(env, supers, exact)
            else:
                self.rout = constraints.fromVariable(env, self.rval)
        else:
            self.rout = None

    def propagateConstraints(self, *incons):
        if self.instruction[0] != 'invokestatic' and incons[0].isConstNull():
            return throw(self.eout_npe)
        return returnOrThrow(self.rout, self.eout)

# TODO - cleanup
class InvokeDynamic(BaseOp):
    has_side_effects = True

    def __init__(self, parent, desc, args):
        super(InvokeDynamic, self).__init__(parent, args, makeException=True)
        self.desc = desc
        vtypes = parseMethodDescriptor(self.desc)[1]

        dtype = None
        if vtypes:
            stype = verifierToSSAType(vtypes[0])
            dtype = objtypes.verifierToSynthetic(vtypes[0])
            cat = len(vtypes)
            self.rval = parent.makeVariable(stype, origin=self)
            self.returned = [self.rval] + [None]*(cat-1)
        else:
            self.rval, self.returned = None, []

        # just use a fixed constraint until we can do interprocedural analysis
        # output order is rval, exception, defined by BaseOp.getOutputs
        env = parent.env
        self.eout = ObjectConstraint.fromTops(env, [objtypes.ThrowableTT], [], nonnull=True)
        if self.rval is not None:
            if self.rval.type == SSA_OBJECT:
                supers, exact = objtypes.declTypeToActual(env, dtype)
                self.rout = ObjectConstraint.fromTops(env, supers, exact)
            else:
                self.rout = constraints.fromVariable(env, self.rval)
        else:
            self.rout = None

    def propagateConstraints(self, *incons):
        return returnOrThrow(self.rout, self.eout)



import itertools
import operator

from ..constraints import IntConstraint

def split_pow2ranges(x,y):
    '''split given range into power of two ranges of form [x, x+2^k)'''
    out = []
    while x<=y:
        # The largest power of two range of the form x,k
        # has k min of number of zeros at end of x
        # and the largest power of two that fits in y-x
        bx = bin(x)
        numzeroes = float('inf') if x==0 else (len(bx)-bx.rindex('1')-1)
        k = min(numzeroes, (y-x+1).bit_length()-1)
        out.append((x,k))
        x += 1<<k
    assert x == y+1
    return out

def propagateBitwise(arg1, arg2, op, usemin, usemax):
    ranges1 = split_pow2ranges(arg1.min, arg1.max)
    ranges2 = split_pow2ranges(arg2.min, arg2.max)

    vals = []
    for (s1,k1),(s2,k2) in itertools.product(ranges1, ranges2):
        # there are three parts. The high bits fixed in both arguments,
        # the middle bits fixed in one but not the other, and the
        # lowest bits which can be chosen freely for both arguments
        # high = op(h1,h2) and low goes from 0 to 1... but the range of
        # the middle depends on the particular operation
        # 0-x, x-1 and 0-1 for and, or, and xor respectively
        if k1 > k2:
            (s1,k1),(s2,k2) = (s2,k2),(s1,k1)

        mask1 = (1<<k1) - 1
        mask2 = (1<<k2) - 1 - mask1

        high = op(s1, s2) & ~(mask1 | mask2)
        midmin = (s1 & mask2) if usemin else 0
        midmax = (s1 & mask2) if usemax else mask2

        vals.append(high | midmin)
        vals.append(high | midmax | mask1)
    return IntConstraint.range(arg1.width, min(vals), max(vals))

def propagateAnd(x, y):
    return propagateBitwise(x, y, operator.__and__, False, True)

def propagateOr(x, y):
    return propagateBitwise(x, y, operator.__or__, True, False)

def propagateXor( x, y):
    return propagateBitwise(x, y, operator.__xor__, False, False)



import math

INF_MAG = 1, None
ZERO_MAG = 0, None

# Numbers are represented as (sign, (mantissa, exponent))
# For finite nonzero values, the float value is sign * mantissa * 2 ^ (exponent - mbits - 1)
# Mantissa is normalized to always be within (2 ^ mbits) <= m < (2 ^ mbits + 1) even for subnormal numbers
NAN = None,(None,None)
INF = 1,INF_MAG
NINF = -1,INF_MAG
ZERO = 1,ZERO_MAG
NZERO = -1,ZERO_MAG

# Key suitable for sorting finite (normalized) nonzero values
def _fkey(s, me):
    m, e = me
    return (s,s*e,s*m)

sortkey = _fkey

# Size info for type - mantissa bits, min exponent, max exponent
FLOAT_SIZE = 23,-126,127
DOUBLE_SIZE = 52,-1022,1023

def flog(x):
    '''returns f such that 2**f <= x < 2**(f+1)'''
    assert x > 0
    return len(bin(x))-3

def roundMag(size, mag):
    '''Round (unnormalized) magnitude to nearest representable magnitude with ties going to 0 lsb'''
    mbits, emin, emax = size
    m, e = mag
    assert m >= 1
    f = flog(m)

    if e+f < emin: # subnormal
        dnmin = emin - mbits
        if e+f < (dnmin - 1):
            return ZERO_MAG
        if e > dnmin:
            m = m << (e - dnmin)
            f += (e - dnmin)
            e = dnmin
        s = dnmin - e
        i = m >> s
        r = (m - (i << s)) * 2
        h = 1 << s
        if r > h or r == h and (i&1):
            i += 1
        return i, e+s-mbits-1
    else:
        if f < mbits:
            m = m << (mbits - f)
            f = mbits
        s = f - mbits
        if (e+f) > emax:
            return INF_MAG
        i = m >> s
        r = (m - (i << s)) * 2
        h = 1 << s
        if r > h or r == h and (i&1):
            i += 1
            if i == (1<<mbits):
                i = i >> 1
                e += 1
                if e > emax:
                    return INF_MAG
        return i, e+s-mbits-1

def fromRawFloat(size, x):
    if math.isnan(x):
        return NAN
    sign = int(math.copysign(1, x))
    x = math.copysign(x, 1)

    if math.isinf(x):
        return sign, INF_MAG
    elif x == 0.0:
        return sign, ZERO_MAG
    else:
        m, e = math.frexp(x)
        m = int(m * (1<<(size[0]+1)))
        return sign, roundMag(size, (m, e))



import heapq

from .cfg import flattenDict, makeGraph

# Variables x and y can safely be merged when it is true that for any use of y (respectively x)
# that sees a definition of y, either there are no intervening definitions of x, or x was known
# to be equal to y *at the point of its most recent definition*
# Given this info, we greedily merge related variables, that is, those where one is assigned to the other
# to calculate which variables can be merged, we first have to build a CFG from the Java AST again

class VarInfo(object):
    __slots__ = "key", "defs", "rdefs", "extracount"
    def __init__(self, key):
        self.key = key
        self.defs = set()
        self.rdefs = set()
        self.extracount = 0

    def priority(self):
        return (len(self.defs) + self.extracount), self.key

class EqualityData(object):
    def __init__(self, d=None):
        # Equal values point to a representative object instance. Singletons are not represented at all for efficiency
        # None represents the top value (i.e. this point has not been visited yet)
        self.d = d.copy() if d is not None else None

    def _newval(self): return object()

    def initialize(self): # initialize to bottom value (all variables unequal)
        assert self.d is None
        self.d = {}

    def handleAssign(self, var1, var2=None):
        if var1 == var2:
            return
        if var2 is None:
            if var1 in self.d:
                del self.d[var1]
        else:
            self.d[var1] = self.d.setdefault(var2, self._newval())
            assert self.iseq(var1, var2)

    def iseq(self, var1, var2):
        assert var1 != var2
        return var1 in self.d and var2 in self.d and self.d[var1] is self.d[var2]

    def merge_update(self, other):
        if other.d is None:
            return
        elif self.d is None:
            self.d = other.d.copy()
        else:
            d1, d2 = self.d, other.d
            new = {}
            todo = list(set(d1) & set(d2))
            while todo:
                cur = todo.pop()
                matches = [k for k in todo if d1[k] is d1[cur] and d2[k] is d2[cur]]
                if not matches:
                    continue
                new[cur] = self._newval()
                for k in matches:
                    new[k] = new[cur]
                todo = [k for k in todo if k not in new]
            self.d = new

    def copy(self): return EqualityData(self.d)

    def __eq__(self, other):
        if self.d is None or other.d is None:
            return self.d is other.d
        if self.d == other.d:
            return True
        if set(self.d) != set(other.d):
            return False
        match = {}
        for k in self.d:
            if match.setdefault(self.d[k], other.d[k]) != other.d[k]:
                return False
        return True

    def __ne__(self, other): return not self == other
    def __hash__(self): raise TypeError('unhashable type')

def calcEqualityData(graph):
    graph.simplify()
    blocks = graph.blocks
    d = {b:[EqualityData()] for b in blocks}

    d[graph.entry][0].initialize()
    stack = [graph.entry]
    dirty = set(blocks)

    while stack:
        block = stack.pop()
        if block not in dirty:
            continue
        dirty.remove(block)

        cur = d[block][0].copy()
        e_out = EqualityData()
        del d[block][1:]

        for line_t, data in block.lines:
            if line_t == 'def':
                cur.handleAssign(*data)
                d[block].append(cur.copy())
            elif line_t == 'canthrow':
                e_out.merge_update(cur)

        for out, successors in [(e_out, block.e_successors), (cur, block.n_successors)]:
            stack += successors
            for suc in successors:
                old = d[suc][0].copy()
                d[suc][0].merge_update(out)
                if old != d[suc][0]:
                    dirty.add(suc)

    for block in blocks:
        assert d[block][0].d is not None
    assert not dirty
    return d

class VarMergeInfo(object):
    def __init__(self, graph, methodparams, isstatic):
        self.info = {}
        self.final, self.unmergeable, self.external = set(), set(), set()
        self.equality = None # to be calculated later
        self.graph = graph

        self.pending_graph_replaces = {}
        self.touched_vars = set()

        # initialize variables and assignment data
        for var in methodparams:
            self._addvar(var)
        self.external.update(methodparams)
        if not isstatic:
            self.final.add(methodparams[0])

        for block in graph.blocks:
            for line_t, data in block.lines:
                if line_t == 'def':
                    self._addassign(data[0], data[1])
            for caught in block.caught_excepts:
                self._addvar(caught)
                self.external.add(caught)
                self.unmergeable.add(caught)

    # initialization helper funcs
    def _addvar(self, v):
        return self.info.setdefault(v, VarInfo(len(self.info)))

    def _addassign(self, v1, v2):
        info = self._addvar(v1)
        if v2 is not None:
            info.defs.add(v2)
            self._addvar(v2).rdefs.add(v1)
        else:
            info.extracount += 1

    # process helper funcs
    def iseq(self, block, index, v1, v2):
        return self.equality[block][index].iseq(v1, v2)

    def _doGraphReplacements(self):
        self.graph.replace(self.pending_graph_replaces)
        self.pending_graph_replaces = {}
        self.touched_vars = set()

    def compat(self, v1, v2, doeq):
        if v1 in self.touched_vars or v2 in self.touched_vars:
            self._doGraphReplacements()

        blocks = self.graph.blocks
        vok = {b:3 for b in blocks} # use bitmask v1ok = 1<<0, v2ok = 1<<1

        stack = [b for b in blocks if v1 in b.vars or v2 in b.vars]
        while stack:
            block = stack.pop()
            cur = vok[block]
            e_out = 3

            if v1 in block.vars or v2 in block.vars:
                defcount = 0
                for line_t, data in block.lines:
                    if line_t == 'use':
                        if (data == v1 and not cur & 1) or (data == v2 and not cur & 2):
                            return False
                    elif line_t == 'def':
                        defcount += 1

                        if data[0] == v1 and data[1] != v1:
                            cur = 1
                        elif data[0] == v2 and data[1] != v2:
                            cur = 2
                        if doeq and self.iseq(block, defcount, v1, v2):
                            cur = 3
                    elif line_t == 'canthrow':
                        e_out &= cur
            else:
                # v1 and v2 not touched in this block, so there is nothing to do
                e_out = cur

            for out, successors in [(e_out, block.e_successors), (cur, block.n_successors)]:
                for suc in successors:
                    if vok[suc] & out != vok[suc]:
                        stack.append(suc)
                        vok[suc] &= out
        return True

    def process(self, replace, doeq):
        final, unmergeable, external = self.final, self.unmergeable, self.external
        d = self.info
        work_q = [(info.priority(), var) for var, info in d.items()]
        heapq.heapify(work_q)
        dirty = set(d) - external

        while work_q:
            _, cur = heapq.heappop(work_q)
            if (cur in external) or cur not in dirty:
                continue
            dirty.remove(cur)

            candidate_set = d[cur].defs - unmergeable
            if len(d[cur].defs) > 1 or d[cur].extracount > 0:
                candidate_set = candidate_set - final
            candidates = [v for v in candidate_set if v.dtype == cur.dtype]
            candidates = sorted(candidates, key=lambda v:d[v].key)
            assert cur not in candidates

            # find first candidate that is actually compatible
            for parent in candidates:
                if self.compat(cur, parent, doeq):
                    break
            else:
                continue # no candidates found

            replace[cur] = parent
            self.pending_graph_replaces[cur] = parent
            self.touched_vars.add(cur)
            self.touched_vars.add(parent)

            infc, infp = d[cur], d[parent]
            # Be careful, there could be a loop with cur in parent.defs
            infc.defs.remove(parent)
            infc.rdefs.discard(parent)
            infp.rdefs.remove(cur)
            infp.defs.discard(cur)

            for var in d[cur].rdefs:
                d[var].defs.remove(cur)
                d[var].defs.add(parent)
                heapq.heappush(work_q, (d[var].priority(), var))

            for var in d[cur].defs:
                d[var].rdefs.remove(cur)
                d[var].rdefs.add(parent)

            d[parent].defs |= d[cur].defs
            d[parent].rdefs |= d[cur].rdefs
            d[parent].extracount += d[cur].extracount

            del d[cur]
            heapq.heappush(work_q, (d[parent].priority(), parent))
            dirty.add(parent)

    def processMain(self, replace):
        self.process(replace, False)
        self._doGraphReplacements()
        self.equality = calcEqualityData(self.graph)
        self.process(replace, True)

###############################################################################
def mergeVariables(root, isstatic, parameters):
    # first, create CFG from the Java AST
    graph = makeGraph(root)
    mergeinfo = VarMergeInfo(graph, parameters, isstatic)

    replace = {}
    mergeinfo.processMain(replace)

    flattenDict(replace)
    return replace



DIRECTIVE = r'\.[a-z]+'
WORD = r'(?:[a-zA-Z_$\(<]|\[[A-Z\[])[\w$;\/\[\(\)<>*+-]*'
FOLLOWED_BY_WHITESPACE = r'(?=\s|\Z)'
REF = r'\[[a-z0-9_:]+\]'
LABEL_DEF = r'L\w+:'

COMMENT = r';.*'
# Match optional comment and at least one newline, followed by any number of empty/whitespace lines
NEWLINES = r'(?:{})?\n\s*'.format(COMMENT)

HEX_DIGIT = r'[0-9a-fA-F]'
ESCAPE_SEQUENCE = r'''\\(?:U00(?:10|0{hd}){hd}{{4}}|u{hd}{{4}}|x{hd}{{2}}|[btnfr'"\\0-7])'''.format(hd=HEX_DIGIT)
# See http://stackoverflow.com/questions/430759/regex-for-managing-escaped-characters-for-items-like-string-literals/5455705# 5455705
STRING_LITERAL = r'''
[bB]?(?:
    "
        [^"\n\\]*               # any number of unescaped characters
        (?:{es}[^"\n\\]*       # escape sequence followed by 0 or more unescaped
        )*
    "
|
    '
        [^'\n\\]*               # any number of unescaped characters
        (?:{es}[^'\n\\]*       # escape sequence followed by 0 or more unescaped
        )*
    '
)'''.format(es=ESCAPE_SEQUENCE)
# For error detection
STRING_START = r'''[bB]?(?:"(?:[^"\\\n]|{es})*|'(?:[^'\\\n]|{es})*)'''.format(es=ESCAPE_SEQUENCE)

# Careful here: | is not greedy so hex must come first
INT_LITERAL = r'[+-]?(?:0[xX]{hd}+|[1-9][0-9]*|0)[lL]?'.format(hd=HEX_DIGIT)
FLOAT_LITERAL = r'''(?:
    (?:
        [-+][Ii][Nn][Ff][Ii][Nn][Ii][Tt][Yy]|           # Nan and Inf both have mandatory sign
        [-+][Nn][Aa][Nn]
            (?:<0[xX]{hd}+>)?                           # Optional suffix for nonstandard NaNs
    )|
    [-+]?(?:
        \d+\.\d+(?:[eE][+-]?\d+)?|                         # decimal float
        \d+[eE][+-]?\d+|                                   # decimal float with no fraction (exponent mandatory)
        0[xX]{hd}+(?:\.{hd}+)?[pP][+-]?\d+                 # hexidecimal float
        )
    )[fF]?
'''.format(hd=HEX_DIGIT)



from __future__ import print_function

import ast
import struct
import sys

from ..classfileformat import mutf8
from ..util.thunk import thunk

from . import assembly, codes, pool
from .flags import FLAGS
from .instructions import OPNAME_TO_BYTE, OP_CLS, OP_FMIM_TO_GUESS, OP_LBL, OP_NONE, OP_SHORT
from .tokenize import AsssemblerError, Tokenizer
from .writer import Writer

def unique(vals):
    return len(vals) == len(set(vals))

def formatList(vals):
    if len(vals) > 1:
        vals[-1] = 'or ' + vals[-1]
    sep = ', ' if len(vals) > 2 else ' '
    return sep.join(vals)

def parseFloat(s, isfloat):
    if s.endswith('>'):
        return int(s.partition('<')[-1][:-1], 16)

    if s.lstrip('+-')[:2].lower() == '0x':
        f = float.fromhex(s)
    else:
        f = float(s)

    if isfloat:
        return struct.unpack('>I', struct.pack('>f', f))[0]
    else:
        return struct.unpack('>Q', struct.pack('>d', f))[0]

class Parser(object):
    def __init__(self, tokenizer):
        self.tokenizer = tokenizer
        self.tok = None
        self.consume()

        self.cls = assembly.Class()
        self.field = self.method = self.code = None

    def _next_token(self):
        return self.tokenizer.next()

    def _format_error_args(self, message, tok):
        if tok.type == 'NEWLINES' or tok.type == 'EOF':
            return message, tok.pos, tok.pos+1
        else:
            return message, tok.pos, tok.pos + len(tok.val)

    def error(self, *args):
        messages = args[0::2]
        tokens = args[1::2]
        assert len(messages) == len(tokens)
        self.tokenizer.error(*map(self._format_error_args, messages, tokens))

    def fail(self):
        assert unique(self.triedvals)
        assert unique(self.triedtypes)
        expected = sorted(self.triedtypes) + sorted(map(repr, self.triedvals))
        assert expected
        self.error('Expected {}.'.format(formatList(expected)), self.tok)

    def consume(self):
        tok = self.tok
        self.triedvals = []
        self.triedtypes = []
        self.tok = self._next_token()
        return tok

    def hasv(self, val):
        self.triedvals.append(val)
        return self.tok.val == val

    def hasany(self, vals):
        self.triedvals.extend(vals)
        return self.tok.val in vals

    def hastype(self, t):
        self.triedtypes.append(t)
        return self.tok.type == t

    def asserttype(self, t):
        self.triedtypes.append(t)
        if self.tok.type != t:
            self.fail()

    def tryv(self, val):
        if self.hasv(val):
            self.consume()
            return True
        return False

    def val(self, val):
        if not self.tryv(val):
            self.fail()

    def ateol(self): return self.hastype('NEWLINES')
    def atendtok(self): return self.hasv('.end')

    def listu8(a, w, end, callback):
        count, pos = 0, w.ph8()
        while not end():
            if count >= 255:
                a.error('Maximum 255 items.', a.tok)
            count += 1
            callback(w)
        w.setph8(pos, count)

    def list(a, w, end, callback):
        count, pos = 0, w.ph16()
        while not end():
            if count >= 65535:
                a.error('Maximum 65535 items.', a.tok)
            count += 1
            callback(w)
        w.setph16(pos, count)

    ###########################################################################
    def eol(a): a.asserttype('NEWLINES'), a.consume()

    def _rawint(a):
        a.asserttype('INT_LITERAL')
        return a.tok, ast.literal_eval(a.consume().val.lstrip('+'))

    def boundedint(a, lower, upper):
        tok, x = a._rawint()
        if not lower <= x < upper:
            a.error('Value must be in range {} <= x < {}.'.format(lower, upper), tok)
        return x

    def u8(a): return a.boundedint(0, 1<<8)
    def u16(a): return a.boundedint(0, 1<<16)
    def u32(a): return a.boundedint(0, 1<<32)
    def s8(a): return a.boundedint(-1<<7, 1<<7)
    def s16(a): return a.boundedint(-1<<15, 1<<15)
    def s32(a): return a.boundedint(-1<<31, 1<<31)

    def string(a, maxlen=65535):
        a.asserttype('STRING_LITERAL')
        tok = a.consume()
        tokval = tok.val
        if not tokval[0] in 'bB':
            tokval = 'u' + tokval
        val = ast.literal_eval(tokval)
        if not isinstance(val, bytes):
            val = mutf8.encode(val)
        if len(val) > maxlen:
            a.error('Maximum string length here is {} bytes ({} found).'.format(maxlen, len(val)), tok)
        return val

    def word(a, maxlen=65535):
        a.asserttype('WORD')
        tok = a.consume()
        val = tok.val.encode('ascii')
        if len(val) > maxlen:
            a.error('Maximum identifier length is {} bytes ({} found).'.format(maxlen, len(val)), tok)
        return val

    def identifier(a):
        if a.hastype('WORD'):
            return a.word()
        elif a.hastype('STRING_LITERAL'):
            return a.string()
        a.fail()

    def intl(a):
        a.asserttype('INT_LITERAL')
        tok = a.consume()
        x = ast.literal_eval(tok.val.lstrip('+'))
        if not -1<<31 <= x < 1<<31:
            a.error('Value does not fit into int type.', tok)
        return x % (1 << 32)

    def longl(a):
        a.asserttype('LONG_LITERAL')
        tok = a.consume()
        x = ast.literal_eval(tok.val.lstrip('+').rstrip('lL'))
        if not -1 << 63 <= x < 1 << 63:
            a.error('Value does not fit into long type.', tok)
        return x % (1 << 64)

    def floatl(a):
        a.asserttype('FLOAT_LITERAL')
        return parseFloat(a.consume().val.rstrip('fF'), True)

    def doublel(a):
        a.asserttype('DOUBLE_LITERAL')
        return parseFloat(a.consume().val, False)

    def ref(a, isbs=False):
        a.asserttype('REF')
        tok = a.consume()

        content = tok.val[1:-1]
        bootstrap = content.startswith('bs:')
        if isbs and not bootstrap:
            a.error('Expected bootstrap reference, found constant pool reference.', tok)
        elif not isbs and bootstrap:
            a.error('Expected constant pool reference, found bootstrap reference.', tok)

        val = content.rpartition(':')[2]
        try:
            index = int(val)
            if not 0 <= index < 0xFFFF: # note: strict upper bound
                a.error('Reference must be in range 0 <= x < 65535.', tok)
            return pool.Ref(tok, index=index, isbs=bootstrap)
        except ValueError:
            return pool.Ref(tok, symbol=val, isbs=bootstrap)

    def utfref(a):
        if a.hastype('REF'):
            return a.ref()
        return pool.utf(a.tok, a.identifier())

    def clsref(a, tag='Class'):
        assert tag in 'Class Module Package'.split()
        if a.hastype('REF'):
            return a.ref()
        return pool.single(tag, a.tok, a.identifier())

    def natref(a):
        if a.hastype('REF'):
            return a.ref()
        name = pool.utf(a.tok, a.identifier())
        desc = a.utfref()
        return pool.Ref(name.tok, type='NameAndType', refs=[name, desc])

    def fmimref(a, typeguess):
        # This rule requires extra lookahead
        if a.hastype('REF'):
            first = a.ref()
            # Legacy method syntax
            if a.hastype('WORD'):
                return pool.Ref(first.tok, type=typeguess, refs=[first, a.natref()]) # Krakatau v0
            return first

        elif a.hasany(['Field', 'Method', 'InterfaceMethod']):
            return a.tagged_const()

        # Legacy method syntax - attempt to support Jasmin's awful syntax too
        words = []
        while len(words) < 3 and a.hastype('WORD'):
            words.append((a.tok, a.word()))

        if 1 <= len(words) <= 2 and a.hastype('REF'): # Krakatau v0
            cls = pool.single('Class', *words[0])
            if len(words) == 2:
                name = pool.utf(*words[1])
                return pool.Ref(cls.tok, type=typeguess, refs=[cls, pool.nat(name, a.utfref())])
            return pool.Ref(cls.tok, type=typeguess, refs=[cls, a.natref()])

        if len(words) == 3: # Krakatau v0
            cls = pool.single('Class', *words[0])
            name = pool.utf(*words[1])
            desc = pool.utf(*words[2])
        elif len(words) == 2: # Jasmin field syntax
            tok, cnn = words[0]
            left, _, right = cnn.rpartition(b'/')

            cls = pool.single('Class', tok, left)
            name = pool.utf(tok, right)
            desc = pool.utf(*words[1])
        elif len(words) == 1: # Jasmin method syntax
            tok, cnnd = words[0]
            cnn, _, d = cnnd.partition(b'(')
            left, _, right = cnn.rpartition(b'/')

            cls = pool.single('Class', tok, left)
            name = pool.utf(tok, right)
            desc = pool.utf(tok, b'(' + d)
        else:
            a.fail()
        return pool.Ref(cls.tok, type=typeguess, refs=[cls, pool.nat(name, desc)])

    def bootstrapargs(a):
        while not a.hasv(':'):
            yield a.ref_or_tagged_const(methodhandle=True)
        a.val(':')

    def bsref(a):
        tok = a.tok
        if a.hastype('REF'):
            return a.ref(isbs=True)
        refs = [a.mhnotref(a.tok)]
        refs.extend(a.bootstrapargs())
        return pool.Ref(tok, type='Bootstrap', refs=refs, isbs=True)

    def mhnotref(a, tok):
        if a.hasany(codes.handle_codes):
            code = codes.handle_codes[a.consume().val]
            return pool.Ref(tok, type='MethodHandle', data=code, refs=[a.ref_or_tagged_const()])
        a.fail()

    def tagged_const(a, methodhandle=False, invokedynamic=False):
        tok = a.tok
        if a.tryv('Utf8'):
            return pool.utf(tok, a.identifier())
        elif a.tryv('Int'):
            return pool.primitive(tok.val, tok, a.intl())
        elif a.tryv('Float'):
            return pool.primitive(tok.val, tok, a.floatl())
        elif a.tryv('Long'):
            return pool.primitive(tok.val, tok, a.longl())
        elif a.tryv('Double'):
            return pool.primitive(tok.val, tok, a.doublel())
        elif a.hasany(['Class', 'String', 'MethodType', 'Module', 'Package']):
            a.consume()
            return pool.Ref(tok, type=tok.val, refs=[a.utfref()])
        elif a.hasany(['Field', 'Method', 'InterfaceMethod']):
            a.consume()
            return pool.Ref(tok, type=tok.val, refs=[a.clsref(), a.natref()])
        elif a.tryv('NameAndType'):
            return pool.Ref(tok, type=tok.val, refs=[a.utfref(), a.utfref()])
        elif methodhandle and a.tryv('MethodHandle'):
            return a.mhnotref(tok)
        elif invokedynamic and a.tryv('InvokeDynamic'):
            return pool.Ref(tok, type=tok.val, refs=[a.bsref(), a.natref()])

        elif a.tryv('Bootstrap'):
            if a.hastype('REF'):
                refs = [a.ref()]
            else:
                refs = [a.mhnotref(a.tok)]
            refs.extend(a.bootstrapargs())
            return pool.Ref(tok, type=tok.val, refs=refs, isbs=True)
        a.fail()

    def ref_or_tagged_const(a, isbs=False, methodhandle=False, invokedynamic=False):
        if a.hastype('REF'):
            ref = a.ref(isbs=isbs)
        else:
            ref = a.tagged_const(methodhandle=methodhandle, invokedynamic=invokedynamic)

        if isbs and not ref.isbs:
            a.error('Expected bootstrap reference, found constant pool reference.', ref.tok)
        elif not isbs and ref.isbs:
            a.error('Expected constant pool reference, found bootstrap reference.', ref.tok)
        return ref

    def ldc_rhs(a):
        tok = a.tok
        if a.hastype('INT_LITERAL'):
            return pool.primitive('Int', tok, a.intl())
        elif a.hastype('FLOAT_LITERAL'):
            return pool.primitive('Float', tok, a.floatl())
        elif a.hastype('LONG_LITERAL'):
            return pool.primitive('Long', tok, a.longl())
        elif a.hastype('DOUBLE_LITERAL'):
            return pool.primitive('Double', tok, a.doublel())
        elif a.hastype('STRING_LITERAL'):
            return pool.single('String', a.tok, a.string())
        return a.ref_or_tagged_const(methodhandle=True)

    def flags(a):
        flags = 0
        while a.hasany(FLAGS):
            flags |= FLAGS[a.consume().val]
        return flags

    def lbl(a):
        a.asserttype('WORD')
        if not a.tok.val.startswith('L'):
            a.error('Labels must start with L.', a.tok)
        if a.code is None:
            a.error('Labels may only be used inside of a Code attribute.', a.tok)
        return assembly.Label(a.tok, a.consume().val)

    ###########################################################################
    ### Top level stuff (class, const defs, fields, methods) ##################
    def parseClass(a):
        a.version_opt()
        a.class_start()

        # Workaround for legacy code without .end class
        while not (a.atendtok() or a.hastype('EOF')):
            a.class_item()

        if a.tryv('.end'):
            a.val('class')
            a.asserttype('NEWLINES')
        return a.cls.assemble(a.error)

    def version_opt(a):
        if a.tryv('.version'):
            a.cls.version = a.u16(), a.u16()
            a.cls.useshortcodeattrs = a.cls.version < (45, 3)
            a.eol()

    def class_start(a):
        a.val('.class')
        a.cls.access = a.flags()
        a.cls.this = a.clsref()
        a.eol()

        a.val('.super')
        a.cls.super = a.clsref()
        a.eol()

        while a.tryv('.implements'):
            a.cls.interfaces.append(a.clsref())
            a.eol()

    def class_item(a):
        a.try_const_def() or a.try_field() or a.try_method() or a.try_attribute(a.cls) or a.fail()

    def try_const_def(a):
        if a.hasany(['.const', '.bootstrap']):
            isbs = a.consume().val == '.bootstrap'
            lhs = a.ref(isbs)
            if lhs.isbs != isbs:
                a.error('Const/Bootstrap reference mismatch.', lhs.tok)
            a.val('=')

            rhs = a.ref_or_tagged_const(isbs, methodhandle=True, invokedynamic=True)
            if lhs.israw() and (rhs.israw() or rhs.issym()):
                a.error('Raw references cannot be aliased to another reference.', rhs.tok)
            a.eol()

            a.cls.pool.sub(lhs).adddef(lhs, rhs, a.error)
            return True
        return False

    def try_field(a):
        if a.hasv('.field'):
            f = a.field_start()
            a.initial_value_opt()

            if a.tryv('.fieldattributes'):
                a.eol()
                while not a.atendtok():
                    a.try_attribute(f) or a.fail()
                a.val('.end'), a.val('fieldattributes')

            a.eol()
            a.cls.fields.append(f)
            a.field = None
            return True
        return False

    def field_start(a):
        tok, flags, name, desc = a.consume(), a.flags(), a.utfref(), a.utfref()
        a.field = f = assembly.Field(tok, flags, name, desc)
        return f

    def initial_value_opt(a):
        tok = a.tok
        if a.tryv('='):
            attr = assembly.Attribute(tok, b'ConstantValue')
            attr.data.ref(a.ldc_rhs())
            a.field.attributes.append(attr)

    def try_method(a):
        if a.hasv('.method'):
            m = a.method_start()

            # Legacy syntax
            if a.hasv('.throws'):
                m.attributes.append(assembly.Attribute(a.consume(), b'Exceptions'))
                m.attributes[-1].data.u16(1)
                m.attributes[-1].data.ref(a.clsref())
                a.eol()

            if a.hasv('.limit'):
                a.legacy_method_body()
            else:
                while not a.atendtok():
                    a.try_attribute(m) or a.fail()

            a.val('.end'), a.val('method'), a.eol()
            a.cls.methods.append(m)
            a.method = None
            return True
        return False

    def method_start(a):
        tok, flags, name, _, desc, _ = a.consume(), a.flags(), a.utfref(), a.val(':'), a.utfref(), a.eol()
        a.method = m = assembly.Method(tok, flags, name, desc)
        return m

    def legacy_method_body(a):
        a.code = c = assembly.Code(a.tok, a.cls.useshortcodeattrs)
        limitfunc = a.u8 if c.short else a.u16
        while a.tryv('.limit'):
            if a.tryv('stack'):
                c.stack = limitfunc()
            elif a.tryv('locals'):
                c.locals = limitfunc()
            else:
                a.fail()
            a.eol()
        a.code_body()
        a.code = None

        attr = assembly.Attribute(c.tok, b'Code')
        c.assembleNoCP(attr.data, a.error)
        a.method.attributes.append(attr)

    ###########################################################################
    ### Bytecode ##############################################################
    def code_body(a):
        while a.try_instruction_line() or a.try_code_directive():
            pass
        while not a.atendtok():
            a.try_attribute(a.code) or a.fail()

    def try_instruction_line(a):
        haslbl = a.hastype('LABEL_DEF')
        if haslbl:
            lbl = assembly.Label(a.tok, a.consume().val.rstrip(':'))
            a.code.labeldef(lbl, a.error)

        hasinstr = a.try_instruction()
        if haslbl or hasinstr:
            a.eol()
            return True
        return False

    def try_instruction(a):
        w = a.code.bytecode
        starttok = a.tok
        op = a.tok.val

        if a.hasany(OP_NONE):
            w.u8(OPNAME_TO_BYTE[a.consume().val])
        elif a.hasany(OP_LBL):
            pos = w.pos
            w.u8(OPNAME_TO_BYTE[a.consume().val])

            dtype = 's32' if op.endswith('_w') else 's16'
            w.lbl(a.lbl(), pos, dtype)
        elif a.hasany(OP_SHORT):
            w.u8(OPNAME_TO_BYTE[a.consume().val]), w.u8(a.u8())
        elif a.hasany(OP_CLS):
            w.u8(OPNAME_TO_BYTE[a.consume().val]), w.ref(a.clsref())
        elif a.hasany(OP_FMIM_TO_GUESS):
            w.u8(OPNAME_TO_BYTE[a.consume().val]), w.ref(a.fmimref(OP_FMIM_TO_GUESS[op]))
        elif a.hasv('invokeinterface'):
            w.u8(OPNAME_TO_BYTE[a.consume().val])
            ref = a.fmimref('InterfaceMethod')
            w.ref(ref)

            if a.hastype('INT_LITERAL'):
                w.u8(a.u8()),
            else:
                a.asserttype('NEWLINES') # print more helpful error for malformed refs
                if ref.israw() or ref.issym():
                    a.error('Method descriptor must be specified inline when argument count is omitted.', ref.tok)
                ref = ref.refs[1] # NAT
                if ref.israw() or ref.issym():
                    a.error('Method descriptor must be specified inline when argument count is omitted.', ref.tok)
                ref = ref.refs[1] # utf
                if ref.israw() or ref.issym():
                    a.error('Method descriptor must be specified inline when argument count is omitted.', ref.tok)
                desc = ref.data.lstrip(b'(')
                count = 1
                while desc:
                    if desc.startswith(b'J') or desc.startswith(b'D'):
                        count += 1
                    else:
                        desc = desc.lstrip(b'[')

                    if desc.startswith(b'L'):
                        _, _, desc = desc.partition(b';')
                    elif desc.startswith(b')'):
                        break
                    else:
                        desc = desc[1:]
                    count += 1
                w.u8(count & 255)
            w.u8(0)

        elif a.hasv('invokedynamic'):
            w.u8(OPNAME_TO_BYTE[a.consume().val]), w.ref(a.ref_or_tagged_const(invokedynamic=True)), w.u16(0)
        elif a.hasany(['ldc', 'ldc_w', 'ldc2_w']):
            w.u8(OPNAME_TO_BYTE[a.consume().val])
            rhs = a.ldc_rhs()
            if op == 'ldc':
                if rhs.israw() and rhs.index >= 256:
                    a.error('Ldc index must be <= 255.', rhs.tok)
                w.refu8(rhs)
            else:
                w.ref(rhs)
        elif a.hasv('multianewarray'):
            w.u8(OPNAME_TO_BYTE[a.consume().val]), w.ref(a.clsref()), w.u8(a.u8())
        elif a.hasv('bipush'):
            w.u8(OPNAME_TO_BYTE[a.consume().val]), w.s8(a.s8())
        elif a.hasv('sipush'):
            w.u8(OPNAME_TO_BYTE[a.consume().val]), w.s16(a.s16())
        elif a.hasv('iinc'):
            w.u8(OPNAME_TO_BYTE[a.consume().val]), w.u8(a.u8()), w.s8(a.s8())
        elif a.hasv('wide'):
            w.u8(OPNAME_TO_BYTE[a.consume().val])
            if a.hasv('iinc'):
                w.u8(OPNAME_TO_BYTE[a.consume().val]), w.u16(a.u16()), w.s16(a.s16())
            elif a.hasany(OP_SHORT):
                w.u8(OPNAME_TO_BYTE[a.consume().val]), w.u16(a.u16())
            else:
                a.fail()
        elif a.hasv('newarray'):
            w.u8(OPNAME_TO_BYTE[a.consume().val])
            if a.hasany(codes.newarr_codes):
                w.u8(codes.newarr_codes[a.consume().val])
            else:
                a.fail()
        elif a.hasv('tableswitch'):
            pos = w.pos
            w.u8(OPNAME_TO_BYTE[a.consume().val]), w.writeBytes(b'\0' * ((3-pos) % 4))
            low = a.s32()
            a.eol()

            jumps = []
            while not a.hasv('default'):
                jumps.append(a.lbl()), a.eol()
                if low + len(jumps) - 1 > (1 << 31) - 1:
                    a.error('Table switch index must be at most 2147483647.', jumps[-1].tok)
            if not jumps:
                a.error('Table switch must have at least one non-default jump.', a.tok)

            _, _, default = a.val('default'), a.val(':'), a.lbl()
            w.lbl(default, pos, 's32')
            w.s32(low)
            w.s32(low + len(jumps) - 1)
            for lbl in jumps:
                w.lbl(lbl, pos, 's32')
        elif a.hasv('lookupswitch'):
            pos = w.pos
            w.u8(OPNAME_TO_BYTE[a.consume().val]), w.writeBytes(b'\0' * ((3-pos) % 4))
            a.eol()

            jumps = {}
            prevtoks = {}
            while not a.hasv('default'):
                keytok, key, _, jump, _ = a.tok, a.s32(), a.val(':'), a.lbl(), a.eol()
                if key in jumps:
                    a.error('Duplicate lookupswitch key.', keytok,
                            'Key previously defined here:', prevtoks[key])
                elif len(jumps) > 1<<31 - 1:
                    a.error('Lookup switch can have at most 2147483647 jumps.', keytok)
                jumps[key] = jump
                prevtoks[key] = keytok

            _, _, default = a.val('default'), a.val(':'), a.lbl()
            w.lbl(default, pos, 's32')
            w.s32(len(jumps))
            for key in sorted(jumps):
                w.s32(key)
                w.lbl(jumps[key], pos, 's32')
        else:
            return False

        if len(w) > a.code.maxcodelen:
            self.error('Maximum bytecode length is {} (current {}).'.format(a.code.maxcodelen, len(w)), starttok)
        return True

    def try_code_directive(a):
        tok = a.tok
        if a.tryv('.catch'):
            if a.code.exceptcount + 1 > 0xFFFF:
                a.error('Maximum 65535 exception handlers per method.', tok)
            ref, (froml, tol), _, usingl, _ = a.clsref(), a.code_range(), a.val('using'), a.lbl(), a.eol()
            a.code.catch(ref, froml, tol, usingl)
            return True
        elif a.tryv('.stack'):
            w = a.code.stackdata
            pos = a.code.bytecode.pos
            delta_offset = pos - a.code.laststackoff - 1
            frame_type = a.tok.val
            if delta_offset < 0:
                a.error('Stack frame has same offset as previous frame.', tok)

            if a.tryv('same'):
                a._check_delta(tok, frame_type, delta_offset, 63)
                w.u8(delta_offset)
            elif a.tryv('stack_1'):
                a._check_delta(tok, frame_type, delta_offset, 63)
                w.u8(delta_offset + 64)
                a.verification_type(w)
            elif a.tryv('stack_1_extended'):
                a._check_delta(tok, frame_type, delta_offset, 0xFFFF)
                w.u8(247)
                w.u16(delta_offset)
                a.verification_type(w)
            elif a.tryv('chop'):
                a._check_delta(tok, frame_type, delta_offset, 0xFFFF)
                w.u8(251 - a.boundedint(1, 4))
                w.u16(delta_offset)
            elif a.tryv('same_extended'):
                a._check_delta(tok, frame_type, delta_offset, 0xFFFF)
                w.u8(251)
                w.u16(delta_offset)
            elif a.tryv('append'):
                a._check_delta(tok, frame_type, delta_offset, 0xFFFF)

                tag = 252
                temp = Writer()
                a.verification_type(temp)
                if not a.ateol():
                    tag += 1
                    a.verification_type(temp)
                    if not a.ateol():
                        tag += 1
                        a.verification_type(temp)

                w.u8(tag)
                w.u16(delta_offset)
                w += temp
            elif a.tryv('full'):
                a._check_delta(tok, frame_type, delta_offset, 0xFFFF)
                w.u8(255)
                w.u16(delta_offset)
                a.eol(), a.val('locals'), a.list(w, a.ateol, a.verification_type)
                a.eol(), a.val('stack'), a.list(w, a.ateol, a.verification_type)
                a.eol(), a.val('.end'), a.val('stack')
            else:
                a.fail()

            a.eol()
            a.code.laststackoff = pos
            a.code.stackcount += 1
            return True
        elif a.tryv('.noimplicitstackmap'):
            a.eol()
            a.code.dont_generate_stackmap = True

        return False

    def code_range(a):
        _, start, _, end = a.val('from'), a.lbl(), a.val('to'), a.lbl()
        return start, end

    def _check_delta(a, tok, frame_type, delta_offset, maxv):
        if delta_offset > maxv:
            a.error('Stack frame type "{}" must appear at most {} bytes after the previous frame (actual offset is {}).'.format(frame_type, maxv+1, delta_offset+1), tok)

    def verification_type(a, w):
        val = a.tok.val
        if not a.hasany(codes.vt_codes):
            a.fail()

        w.u8(codes.vt_codes[a.consume().val])
        if val == 'Object':
            w.ref(a.clsref())
        elif val == 'Uninitialized':
            w.lbl(a.lbl(), 0, 'u16')

    ###########################################################################
    ### Attributes ############################################################
    def try_attribute(a, parent):
        if a.hasv('.attribute'):
            startok, name = a.consume(), a.utfref()
            if a.tryv('length'):
                attr = assembly.Attribute(startok, name, length=a.u32())
            else:
                attr = assembly.Attribute(startok, name)

            # Now get data
            if a.hastype('STRING_LITERAL'):
                attr.data.writeBytes(a.string(maxlen=0xFFFFFFFF))
            else:
                namedattr = a.maybe_named_attribute(attr)
                if namedattr is not None:
                    attr.data = namedattr.data
                else:
                    a.fail()
            a.eol()
            parent.attributes.append(attr)
            return True
        else:
            namedattr = a.maybe_named_attribute(None)
            if namedattr is not None:
                a.eol()
                parent.attributes.append(namedattr)
                return True
        return False

    def maybe_named_attribute(a, wrapper_attr):
        starttok = a.tok
        def create(name):
            attr = assembly.Attribute(starttok, name)
            return attr, attr.data

        if a.tryv('.annotationdefault'):
            attr, w = create(b'AnnotationDefault')
            a.element_value(w)
        elif a.tryv('.bootstrapmethods'):
            attr, w = create(b'BootstrapMethods')
            a.cls.bootstrapmethods = wrapper_attr or attr
        elif a.code is None and a.tryv('.code'):
            a.code = c = assembly.Code(starttok, a.cls.useshortcodeattrs)
            limitfunc = a.u8 if c.short else a.u16
            _, c.stack, _, c.locals, _ = a.val('stack'), limitfunc(), a.val('locals'), limitfunc(), a.eol()
            a.code_body()
            a.val('.end'), a.val('code')
            a.code = None
            attr, w = create(b'Code')
            c.assembleNoCP(w, a.error)
        elif a.tryv('.constantvalue'):
            attr, w = create(b'ConstantValue')
            w.ref(a.ldc_rhs())
        elif a.tryv('.deprecated'):
            attr, w = create(b'Deprecated')
        elif a.tryv('.enclosing'):
            attr, w = create(b'EnclosingMethod')
            a.val('method'), w.ref(a.clsref()), w.ref(a.natref())
        elif a.tryv('.exceptions'):
            attr, w = create(b'Exceptions')
            a.list(w, a.ateol, a._class_item)
        elif a.tryv('.innerclasses'):
            attr, w = create(b'InnerClasses')
            a.eol(), a.list(w, a.atendtok, a._innerclasses_item), a.val('.end'), a.val('innerclasses')
        elif a.tryv('.linenumbertable'):
            attr, w = create(b'LineNumberTable')
            a.eol(), a.list(w, a.atendtok, a._linenumber_item), a.val('.end'), a.val('linenumbertable')
        elif a.tryv('.localvariabletable'):
            attr, w = create(b'LocalVariableTable')
            a.eol(), a.list(w, a.atendtok, a._localvariabletable_item), a.val('.end'), a.val('localvariabletable')
        elif a.tryv('.localvariabletypetable'):
            attr, w = create(b'LocalVariableTypeTable') # reuse _localvariabletable_item func
            a.eol(), a.list(w, a.atendtok, a._localvariabletable_item), a.val('.end'), a.val('localvariabletypetable')
        elif a.tryv('.methodparameters'):
            attr, w = create(b'MethodParameters')
            a.eol(), a.listu8(w, a.atendtok, a._methodparams_item), a.val('.end'), a.val('methodparameters')
        elif a.tryv('.module'):
            print('Warning! Assembler syntax for Java 9 modules is experimental and subject to change. Please file an issue on Github if you have any opinions or feedback about the syntax')
            attr, w = create(b'Module')
            w.ref(a.utfref()), w.u16(a.flags()), a.consume(), w.ref(a.utfref()), a.eol(),
            a._mod_list(w, '.requires', a._mod_requires_item)
            a._mod_list(w, '.exports', a._mod_exports_item)
            a._mod_list(w, '.opens', a._mod_exports_item)
            a._mod_list(w, '.uses', a._mod_uses_item)
            a._mod_list(w, '.provides', a._mod_provides_item)
            a.val('.end'), a.val('module')

        elif a.tryv('.modulemainclass'):
            print('Warning! Assembler syntax for Java 9 modules is experimental and subject to change. Please file an issue on Github if you have any opinions or feedback about the syntax')
            attr, w = create(b'ModuleMainClass')
            w.ref(a.clsref())
        elif a.tryv('.modulepackages'):
            print('Warning! Assembler syntax for Java 9 modules is experimental and subject to change. Please file an issue on Github if you have any opinions or feedback about the syntax')
            attr, w = create(b'ModulePackages')
            a.list(w, a.ateol, a._package_item)

        elif a.tryv('.runtime'):
            if not a.hasany(['visible', 'invisible']):
                a.fail()
            prefix = b'Runtime' + a.consume().val.capitalize().encode()

            if a.tryv('annotations'):
                attr, w = create(prefix + b'Annotations')
                a.eol(), a.list(w, a.atendtok, a.annotation_line)
            elif a.tryv('paramannotations'):
                attr, w = create(prefix + b'ParameterAnnotations')
                a.eol(), a.listu8(w, a.atendtok, a.param_annotation_line)
            elif a.tryv('typeannotations'):
                attr, w = create(prefix + b'TypeAnnotations')
                a.eol(), a.list(w, a.atendtok, a.type_annotation_line)
            else:
                a.fail()
            a.val('.end'), a.val('runtime')
        elif a.code is not None and a.tryv('.stackmaptable'):
            attr, w = create(b'StackMapTable')
            a.code.stackmaptable = wrapper_attr or attr
        elif a.tryv('.signature'):
            attr, w = create(b'Signature')
            w.ref(a.utfref())
        elif a.tryv('.sourcedebugextension'):
            attr, w = create(b'SourceDebugExtension')
            data = a.string(maxlen=0xFFFFFFFF)
            w.writeBytes(data)
        elif a.tryv('.sourcefile'):
            attr, w = create(b'SourceFile')
            w.ref(a.utfref())
        elif a.tryv('.synthetic'):
            attr, w = create(b'Synthetic')
        else:
            return None
        return attr

    def _class_item(a, w): w.ref(a.clsref())
    def _package_item(a, w): w.ref(a.clsref(tag='Package'))
    def _module_item(a, w): w.ref(a.clsref(tag='Package'))
    def _innerclasses_item(a, w): w.ref(a.clsref()), w.ref(a.clsref()), w.ref(a.utfref()), w.u16(a.flags()), a.eol()
    def _linenumber_item(a, w): w.lbl(a.lbl(), 0, 'u16'), w.u16(a.u16()), a.eol()
    def _methodparams_item(a, w): w.ref(a.utfref()), w.u16(a.flags()), a.eol()

    def _localvariabletable_item(a, w):
        ind, _, name, desc, _range, _ = a.u16(), a.val('is'), a.utfref(), a.utfref(), a.code_range(), a.eol()
        w.lblrange(*_range), w.ref(name), w.ref(desc), w.u16(ind)

    # module attr callbacks
    def _mod_list(a, w, startdir, cb): a.list(w, lambda: not a.tryv(startdir), cb)
    def _mod_requires_item(a, w):
        w.ref(a.clsref(tag='Module')), w.u16(a.flags()), a.val('version'), w.ref(a.utfref()), a.eol()
    def _mod_exports_item(a, w):
        w.ref(a.clsref(tag='Package')), w.u16(a.flags())
        if a.tryv('to'):
            a.list(w, a.ateol, a._module_item)
        else:
            w.u16(0) # count of 0 targets
        a.eol()
    def _mod_uses_item(a, w): w.ref(a.clsref()), a.eol()
    def _mod_provides_item(a, w):
        w.ref(a.clsref()), a.val('with')
        a.list(w, a.ateol, a._class_item)
        a.eol()

    ###########################################################################
    ### Annotations ###########################################################
    def annotation_line(a, w):
        a.val('.annotation'), a.annotation_contents(w), a.val('.end'), a.val('annotation'), a.eol()

    def param_annotation_line(a, w):
        a.val('.paramannotation'), a.eol()
        a.list(w, a.atendtok, a.annotation_line)
        a.val('.end'), a.val('paramannotation'), a.eol()

    def type_annotation_line(a, w):
        a.val('.typeannotation'), a.ta_target_info(w), a.ta_target_path(w)
        a.annotation_contents(w), a.val('.end'), a.val('typeannotation'), a.eol()

    def ta_target_info(a, w):
        w.u8(a.u8())
        if a.tryv('typeparam'):
            w.u8(a.u8())
        elif a.tryv('super'):
            w.u16(a.u16())
        elif a.tryv('typeparambound'):
            w.u8(a.u8()), w.u8(a.u8())
        elif a.tryv('empty'):
            pass
        elif a.tryv('methodparam'):
            w.u8(a.u8())
        elif a.tryv('throws'):
            w.u16(a.u16())
        elif a.tryv('localvar'):
            a.eol()
            a.list(w, a.atendtok, a._localvarrange)
            a.val('.end'), a.val('localvar')
        elif a.tryv('catch'):
            w.u16(a.u16())
        elif a.tryv('offset'):
            w.lbl(a.lbl(), 0, 'u16')
        elif a.tryv('typearg'):
            w.lbl(a.lbl(), 0, 'u16'), w.u8(a.u8())
        else:
            a.fail()
        a.eol()

    def _localvarrange(a, w):
        if a.tryv('nowhere'): # WTF, Java?
            w.u16(0xFFFF), w.u16(0xFFFF)
        else:
            w.lblrange(*a.code_range())
        w.u16(a.u16()), a.eol()

    def ta_target_path(a, w):
        a.val('.typepath'), a.eol()
        a.listu8(w, a.atendtok, a._type_path_segment)
        a.val('.end'), a.val('typepath'), a.eol()

    def _type_path_segment(a, w):
        w.u8(a.u8()), w.u8(a.u8()), a.eol()

    # The following are recursive and can be nested arbitrarily deep,
    # so we use generators and a thunk to avoid the Python stack limit.
    def element_value(a, w): thunk(a._element_value(w))
    def annotation_contents(a, w): thunk(a._annotation_contents(w))

    def _element_value(a, w):
        if not a.hasany(codes.et_tags):
            a.fail()

        tag = a.consume().val
        w.u8(codes.et_tags[tag])
        if tag == 'annotation':
            (yield a._annotation_contents(w)), a.val('.end'), a.val('annotation')
        elif tag == 'array':
            a.eol()
            count, pos = 0, w.ph16()
            while not a.atendtok():
                if count >= 65535:
                    a.error('Maximum 65535 items in annotation array element.', a.tok)
                count += 1
                (yield a._element_value(w)), a.eol()
            w.setph16(pos, count)
            a.val('.end'), a.val('array')
        elif tag == 'enum':
            w.ref(a.utfref()), w.ref(a.utfref())
        elif tag == 'class' or tag == 'string':
            w.ref(a.utfref())
        else:
            w.ref(a.ldc_rhs())

    def _annotation_contents(a, w):
        w.ref(a.utfref()), a.eol()
        count, pos = 0, w.ph16()
        while not a.atendtok():
            if count >= 65535:
                a.error('Maximum 65535 items in annotation.', a.tok)
            count += 1
            w.ref(a.utfref()), a.val('='), (yield a._element_value(w)), a.eol()
        w.setph16(pos, count)

def assemble(source, filename, fatal=False):
    tokenizer = Tokenizer(source, filename)
    try:
        while not tokenizer.atend():
            name, data = Parser(tokenizer).parseClass()
            yield name, data
    except AsssemblerError:
        if fatal:
            raise
        sys.exit(1)
    except Exception:
        if fatal:
            raise
        import traceback
        traceback.print_exc()
        print('If you see this message, please file an issue at https://github.com/Storyyeller/Krakatau/issues, including the error message and the assembly file that caused the error.\n')



from collections import namedtuple as nt

from .. import floatutil as fu
from ..verifier import verifier_types as vtypes

slots_t = nt('slots_t', ('locals', 'stack'))

def _localsAsList(self): return [t[1] for t in sorted(self.locals.items())]
slots_t.localsAsList = property(_localsAsList)

# types
SSA_INT = 'int', 32
SSA_LONG = 'int', 64
SSA_FLOAT = 'float', fu.FLOAT_SIZE
SSA_DOUBLE = 'float', fu.DOUBLE_SIZE
SSA_OBJECT = 'obj',

def verifierToSSAType(vtype):
    vtype_dict = {vtypes.T_INT:SSA_INT,
                vtypes.T_LONG:SSA_LONG,
                vtypes.T_FLOAT:SSA_FLOAT,
                vtypes.T_DOUBLE:SSA_DOUBLE}
    # These should never be passed in here
    assert vtype.tag not in ('.new','.init')
    vtype = vtypes.withNoConst(vtype)
    if vtypes.objOrArray(vtype):
        return SSA_OBJECT
    elif vtype in vtype_dict:
        return vtype_dict[vtype]
    return None

# Note: This is actually an Extended Basic Block. A normal basic block has to end whenever there is
# an instruction that can throw. This means that there is a seperate basic block for every throwing
# method, which causes horrible performance, especially in a large method with otherwise linear code.
# The solution is to use extended basic blocks, which are like normal basic blocks except that they
# can contain multiple throwing instructions as long as every throwing instruction has the same
# handlers. Due to the use of SSA, we also require that there are no changes to the locals between the
# first and last throwing instruction.
class BasicBlock(object):
    __slots__ = "key phis lines jump unaryConstraints predecessors inslots throwvars chpairs except_used locals_at_except".split()

    def __init__(self, key):
        self.key = key
        self.phis = None # The list of phi statements merging incoming variables
        self.lines = [] # List of operations in the block
        self.jump = None # The exit point (if, goto, etc)

        # Holds constraints (range and type information) for each variable in the block.
        # If the value is None, this variable cannot be reached
        self.unaryConstraints = None
        # List of predecessor pairs in deterministic order
        self.predecessors = []

        # temp vars used during graph creation
        self.inslots = None
        self.throwvars = []
        self.chpairs = None
        self.except_used = None
        self.locals_at_except = None

    def filterVarConstraints(self, keepvars):
        self.unaryConstraints = {k:v for k,v in self.unaryConstraints.items() if k in keepvars}

    def removePredPair(self, pair):
        self.predecessors.remove(pair)
        for phi in self.phis:
            del phi.dict[pair]

    def replacePredPair(self, oldp, newp):
        self.predecessors[self.predecessors.index(oldp)] = newp
        for phi in self.phis:
            phi.dict[newp] = phi.dict[oldp]
            del phi.dict[oldp]

    def __str__(self):   # pragma: no cover
        return 'Block ' + str(self.key)
    __repr__ = __str__



from ..constraints import IntConstraint, return_

from . import bitwise_util
from .base import BaseOp

class Truncate(BaseOp):
    def __init__(self, parent, arg, signed, width):
        super(Truncate, self).__init__(parent, [arg])

        self.signed, self.width = signed, width
        self.rval = parent.makeVariable(arg.type, origin=self)

    def propagateConstraints(self, x):
        # get range of target type
        w = self.width
        intw = x.width
        assert w < intw
        M = 1<<w

        mask = IntConstraint.const(intw, M-1)
        x = bitwise_util.propagateAnd(x,mask)

        # We have the mods in the range [0,M-1], but we want it in the range
        # [-M/2, M/2-1] so we need to find the new min and max
        if self.signed:
            HM = M>>1

            parts = [(i-M if i>=HM else i) for i in (x.min, x.max)]
            if x.min <= HM-1 <= x.max:
                parts.append(HM-1)
            if x.min <= HM <= x.max:
                parts.append(-HM)

            assert -HM <= min(parts) <= max(parts) <= HM-1
            return return_(IntConstraint.range(intw, min(parts), max(parts)))
        else:
            return return_(x)



from .. import excepttypes
from ..constraints import ObjectConstraint, maybeThrow

from .base import BaseOp

class TryReturn(BaseOp):
    def __init__(self, parent, canthrow=True):
        super(TryReturn, self).__init__(parent, [], makeException=True)
        self.outExceptionCons = ObjectConstraint.fromTops(parent.env, [], (excepttypes.MonState,), nonnull=True)

    def propagateConstraints(self):
        return maybeThrow(self.outExceptionCons)



import copy

from .ssa_types import slots_t

class ProcInfo(object):
    def __init__(self, retblock, target):
        self.retblock = retblock
        self.target = target
        self.jsrblocks = []
        assert target is retblock.jump.target

    def __str__(self):   # pragma: no cover
        return 'Proc{}<{}>'.format(self.target.key, ', '.join(str(b.key) for b in self.jsrblocks))
    __repr__ = __str__

###########################################################################################
class ProcJumpBase(object):
    @property
    def params(self):
        return [v for v in self.input.stack + self.input.localsAsList if v is not None]
        # [v for v in self.input.stack if v] + [v for k, v in sorted(self.input.locals.items()) if v]

    def replaceBlocks(self, blockDict):
        self.target = blockDict.get(self.target, self.target)

    def getExceptSuccessors(self): return ()
    def getSuccessors(self): return self.getNormalSuccessors()
    def getSuccessorPairs(self): return [(x,False) for x in self.getNormalSuccessors()]
    def reduceSuccessors(self, pairsToRemove): return self

class ProcCallOp(ProcJumpBase):
    def __init__(self, target, fallthrough, inslots, outslots):
        self.fallthrough = fallthrough
        self.target = target
        self.input = inslots
        self.output = outslots

        for var in self.output.stack + self.output.locals.values():
            if var is not None:
                assert var.origin is None
                var.origin = self

    # def flatOutput(self): return [v for v in self.output.stack if v] + [v for k, v in sorted(self.output.locals.items()) if v]
    def flatOutput(self): return self.output.stack + self.output.localsAsList

    def getNormalSuccessors(self): return self.fallthrough, self.target

class DummyRet(ProcJumpBase):
    def __init__(self, inslots, target):
        self.target = target
        self.input = inslots

    def replaceVars(self, varDict):
        newstack = [varDict.get(v, v) for v in self.input.stack]
        newlocals = {k: varDict.get(v, v) for k, v in self.input.locals.items()}
        self.input = slots_t(stack=newstack, locals=newlocals)

    def getNormalSuccessors(self): return ()
    def clone(self): return copy.copy(self) # target and input will be replaced later by calls to replaceBlocks/Vars



import collections

from ..constraints import IntConstraint
from ..ssa_types import SSA_INT

from .base import BaseJump
from .goto import Goto
from .ifcmp import If

class Switch(BaseJump):
    def __init__(self, parent, default, table, arguments):
        super(Switch, self).__init__(parent, arguments)

        # get ordered successors since our map will be unordered. Default is always first successor
        if not table:
            ordered = [default]
        else:
            tset = set()
            ordered = [x for x in (default,) + zip(*table)[1] if not x in tset and not tset.add(x)]

        self.successors = ordered
        reverse = collections.defaultdict(set)
        for k,v in table:
            if v != default:
                reverse[v].add(k)
        self.reverse = {k: frozenset(v) for k, v in list(reverse.items())}

    def getNormalSuccessors(self):
        return self.successors

    def replaceBlocks(self, blockDict):
        self.successors = [blockDict.get(key,key) for key in self.successors]
        self.reverse = {blockDict.get(k,k):v for k,v in list(self.reverse.items())}

    def reduceSuccessors(self, pairsToRemove):
        temp = list(self.successors)
        for (child, t) in pairsToRemove:
            temp.remove(child)

        if len(temp) == 0:
            return None
        elif len(temp) == 1:
            return Goto(self.parent, temp.pop())

        if len(temp) < len(self.successors):
            self.successors = temp
            self.reverse = {v:self.reverse[v] for v in temp[1:]}
        return self

    def simplifyToIf(self, block):
        # Try to replace with an if statement if possible
        # e.g. switch(x) {case C: ... default: ...} -> if (x == C) {...} else {...}
        if len(self.successors) == 2:
            cases = self.reverse[self.successors[-1]]
            if len(cases) == 1:
                const = self.parent.makeVariable(SSA_INT)
                const.const = min(cases)
                block.unaryConstraints[const] = IntConstraint.const(32, const.const)
                return If(self.parent, 'eq', self.successors, self.params + [const])
        return self

    ###############################################################################
    def constrainJumps(self, x):
        impossible = []
        for child in self.successors:
            func = self.getSuccessorConstraints((child,False))
            results = func(x)
            if results[0] is None:
                impossible.append((child,False))
        return self.reduceSuccessors(impossible)

    def getSuccessorConstraints(self, xxx_todo_changeme):
        (block, t) = xxx_todo_changeme
        if block in self.reverse:
            cmin = min(self.reverse[block])
            cmax = max(self.reverse[block])
            def propagateConstraints(x):
                if x is None:
                    return None,
                return IntConstraint.range(x.width, max(cmin, x.min), min(cmax, x.max)),
        else:
            allcases = set().union(*list(self.reverse.values()))
            def propagateConstraints(x):
                if x is None or (x.min == x.max and x.min in allcases):
                    return None,
                return x,
        return propagateConstraints



from .. import excepttypes
from ..constraints import ObjectConstraint, maybeThrow

from .base import BaseOp

class Monitor(BaseOp):
    has_side_effects = True

    def __init__(self, parent, args, isExit):
        BaseOp.__init__(self, parent, args, makeException=True)
        self.exit = isExit
        self.env = parent.env

    def propagateConstraints(self, x):
        etypes = ()
        if x.null:
            etypes += (excepttypes.NullPtr,)
        if self.exit and not x.isConstNull():
            etypes += (excepttypes.MonState,)
        eout = ObjectConstraint.fromTops(self.env, [], etypes, nonnull=True)
        return maybeThrow(eout)



from collections import defaultdict as ddict

from .. import graph_util
from ..ssa import objtypes

from . import ast

def flattenDict(replace):
    for k in list(replace):
        while replace[k] in replace:
            replace[k] = replace[replace[k]]

# The basic block in our temporary CFG
# instead of code, it merely contains a list of defs and uses
# This is an extended basic block, i.e. it only terminates in a normal jump(s).
# exceptions can be thrown from various points within the block
class DUBlock(object):
    def __init__(self, key):
        self.key = key
        self.caught_excepts = ()
        self.lines = []     # 3 types of lines: ('use', var), ('def', (var, var2_opt)), or ('canthrow', None)
        self.e_successors = []
        self.n_successors = []
        self.vars = None # vars used or defined within the block. Does NOT include caught exceptions

    def canThrow(self): return ('canthrow', None) in self.lines

    def recalcVars(self):
        self.vars = set()
        for line_t, data in self.lines:
            if line_t == 'use':
                self.vars.add(data)
            elif line_t == 'def':
                self.vars.add(data[0])
                if data[1] is not None:
                    self.vars.add(data[1])

    def replace(self, replace):
        if not self.vars.isdisjoint(replace):
            newlines = []
            for line_t, data in self.lines:
                if line_t == 'use':
                    data = replace.get(data, data)
                elif line_t == 'def':
                    data = replace.get(data[0], data[0]), replace.get(data[1], data[1])
                newlines.append((line_t, data))
            self.lines = newlines
            for k, v in replace.items():
                if k in self.vars:
                    self.vars.remove(k)
                    self.vars.add(v)

    def simplify(self):
        # try to prune redundant instructions
        last = None
        newlines = []
        for line in self.lines:
            if line[0] == 'def':
                if line[1][0] == line[1][1]:
                    continue
            elif line == last:
                continue
            newlines.append(line)
            last = line
        self.lines = newlines
        self.recalcVars()

def varOrNone(expr):
    return expr if isinstance(expr, ast.Local) else None

def canThrow(expr):
    if isinstance(expr, (ast.ArrayAccess, ast.ArrayCreation, ast.Cast, ast.ClassInstanceCreation, ast.FieldAccess, ast.MethodInvocation)):
        return True
    if isinstance(expr, ast.BinaryInfix) and expr.opstr in ('/','%'): # check for possible division by 0
        return expr.dtype not in (objtypes.FloatTT, objtypes.DoubleTT)
    return False

def visitExpr(expr, lines):
    if expr is None:
        return
    if isinstance(expr, ast.Local):
        lines.append(('use', expr))

    if isinstance(expr, ast.Assignment):
        lhs, rhs = map(varOrNone, expr.params)

        # with assignment we need to only visit LHS if it isn't a local in order to avoid spurious uses
        # also, we need to visit RHS before generating the def
        if lhs is None:
            visitExpr(expr.params[0], lines)
        visitExpr(expr.params[1], lines)
        if lhs is not None:
            lines.append(('def', (lhs, rhs)))
    else:
        for param in expr.params:
            visitExpr(param, lines)

    if canThrow(expr):
        lines.append(('canthrow', None))

class DUGraph(object):
    def __init__(self):
        self.blocks = []
        self.entry = None

    def makeBlock(self, key, break_dict, caught_except, myexcept_parents):
        block = DUBlock(key)
        self.blocks.append(block)

        for parent in break_dict[block.key]:
            parent.n_successors.append(block)
        del break_dict[block.key]

        assert (myexcept_parents is None) == (caught_except is None)
        if caught_except is not None: # this is the head of a catch block:
            block.caught_excepts = (caught_except,)
            for parent in myexcept_parents:
                parent.e_successors.append(block)
        return block

    def finishBlock(self, block, catch_stack):
        # register exception handlers for completed old block and calculate var set
        assert(block.vars is None) # make sure it wasn't finished twice
        if block.canThrow():
            for clist in catch_stack:
                clist.append(block)
        block.recalcVars()

    def visitScope(self, scope, break_dict, catch_stack, caught_except=None, myexcept_parents=None, head_block=None):
        # catch_stack is copy on modify
        if head_block is None:
            head_block = block = self.makeBlock(scope.continueKey, break_dict, caught_except, myexcept_parents)
        else:
            block = head_block

        for stmt in scope.statements:
            if isinstance(stmt, (ast.ExpressionStatement, ast.ThrowStatement, ast.ReturnStatement)):
                visitExpr(stmt.expr, block.lines)
                if isinstance(stmt, ast.ThrowStatement):
                    block.lines.append(('canthrow', None))
                continue

            # compound statements
            assert stmt.continueKey is not None
            if isinstance(stmt, (ast.IfStatement, ast.SwitchStatement)):
                visitExpr(stmt.expr, block.lines)

                if isinstance(stmt, ast.SwitchStatement):
                    ft = not stmt.hasDefault()
                else:
                    ft = len(stmt.getScopes()) == 1

                for sub in stmt.getScopes():
                    break_dict[sub.continueKey].append(block)
                    self.visitScope(sub, break_dict, catch_stack)
                if ft:
                    break_dict[stmt.breakKey].append(block)

            elif isinstance(stmt, ast.WhileStatement):
                if stmt.expr != ast.Literal.TRUE: # while(cond)
                    assert stmt.breakKey is not None
                    self.finishBlock(block, catch_stack)
                    block = self.makeBlock(stmt.continueKey, break_dict, None, None)
                    visitExpr(stmt.expr, block.lines)
                    break_dict[stmt.breakKey].append(block)

                break_dict[stmt.continueKey].append(block)
                body_block = self.visitScope(stmt.getScopes()[0], break_dict, catch_stack)
                continue_target = body_block if stmt.expr == ast.Literal.TRUE else block

                for parent in break_dict[stmt.continueKey]:
                    parent.n_successors.append(continue_target)
                del break_dict[stmt.continueKey]

            elif isinstance(stmt, ast.TryStatement):
                new_stack = catch_stack + [[] for _ in stmt.pairs]

                break_dict[stmt.tryb.continueKey].append(block)
                self.visitScope(stmt.tryb, break_dict, new_stack)

                for cdecl, catchb in stmt.pairs:
                    parents = new_stack.pop()
                    self.visitScope(catchb, break_dict, catch_stack, cdecl.local, parents)
                assert new_stack == catch_stack
            else:
                assert isinstance(stmt, ast.StatementBlock)
                break_dict[stmt.continueKey].append(block)
                self.visitScope(stmt, break_dict, catch_stack, head_block=block)

            if not isinstance(stmt, ast.StatementBlock): # if we passed it to subscope, it will be finished in the subcall
                self.finishBlock(block, catch_stack)

            if stmt.breakKey is not None: # start new block after return from compound statement
                block = self.makeBlock(stmt.breakKey, break_dict, None, None)
            else:
                block = None # should never be accessed anyway if we're exiting abruptly

        if scope.jumpKey is not None:
            break_dict[scope.jumpKey].append(block)

        if block is not None:
            self.finishBlock(block, catch_stack)
        return head_block # head needs to be returned in case of loops so we can fix up backedges

    def makeCFG(self, root):
        break_dict = ddict(list)
        self.visitScope(root, break_dict, [])
        self.entry = self.blocks[0] # entry point should always be first block generated

        reached = graph_util.topologicalSort([self.entry], lambda block:(block.n_successors + block.e_successors))
        # if len(reached) != len(self.blocks):
        #     print 'warning, {} blocks unreachable!'.format(len(self.blocks) - len(reached))
        self.blocks = reached

    def replace(self, replace):
        flattenDict(replace)
        for block in self.blocks:
            block.replace(replace)

    def simplify(self):
        for block in self.blocks:
            block.simplify()

def makeGraph(root):
    g = DUGraph()
    g.makeCFG(root)
    return g



import collections
import struct

Label = collections.namedtuple('Label', ['tok', 'sym'])

class Writer(object):
    def __init__(self):
        self.b = bytearray()
        self.refphs = []
        self.refu8phs = []
        self.lblphs = []

        # includes lbl and manual phs but not ref phs
        self._ph8s = set()
        self._ph16s = set()
        self._ph32s = set()

    @property
    def pos(self): return len(self.b)

    def u8(self, x): self.b.append(x)
    def s8(self, x): self.b.append(x % 256)
    def u16(self, x): self.b.extend(struct.pack('>H', x))
    def s16(self, x): self.b.extend(struct.pack('>h', x))
    def u32(self, x): self.b.extend(struct.pack('>I', x))
    def s32(self, x): self.b.extend(struct.pack('>i', x))
    def u64(self, x): self.b.extend(struct.pack('>Q', x))
    def writeBytes(self, b): self.b.extend(b)

    def ref(self, ref):
        self.refphs.append((self.pos, ref))
        self.u16(0)

    def refu8(self, ref):
        self.refu8phs.append((self.pos, ref))
        self.u8(0)

    def ph8(self):
        pos = self.pos
        self.u8(0)
        self._ph8s.add(pos)
        return pos

    def ph16(self):
        pos = self.pos
        self.u16(0)
        self._ph16s.add(pos)
        return pos

    def ph32(self):
        pos = self.pos
        self.u32(0)
        self._ph32s.add(pos)
        return pos

    def lbl(self, lbl, base, dtype):
        pos = self.ph32() if dtype == 's32' else self.ph16()
        self.lblphs.append((pos, lbl, base, dtype))

    def lblrange(self, start, end):
        self.lbl(start, 0, 'u16')
        self.lbl(end, start, 'u16')

    def setph8(self, pos, x):
        assert self.b[pos] == 0
        self.b[pos] = x
        self._ph8s.remove(pos)

    def setph16(self, pos, x):
        assert self.b[pos:pos+2] == b'\0\0'
        self.b[pos:pos+2] = struct.pack('>H', x)
        self._ph16s.remove(pos)

    def setph32(self, pos, x):
        assert self.b[pos:pos+4] == b'\0\0\0\0'
        self.b[pos:pos+4] = struct.pack('>I', x)
        self._ph32s.remove(pos)

    def _getlbl(self, lbl, labels, error):
        if lbl.sym not in labels:
            error('Undefined label', lbl.tok)
        return labels[lbl.sym][1]

    def fillLabels(self, labels, error):
        for pos, lbl, base, dtype in self.lblphs:
            tok = lbl.tok
            lbl = self._getlbl(lbl, labels, error)

            # base can also be a second label
            if isinstance(base, Label):
                base = self._getlbl(base, labels, error)

            offset = lbl - base
            if dtype == 's16':
                if not -1<<15 <= offset < 1<<15:
                    error('Label offset must fit in signed 16 bit int. (offset is {})'.format(offset), tok)
                self.setph16(pos, offset % (1<<16))
            elif dtype == 'u16':
                if not 0 <= offset < 1<<16:
                    error('Label offset must fit in unsigned 16 bit int. (offset is {})'.format(offset), tok)
                self.setph16(pos, offset)
            elif dtype == 's32':
                if not -1<<31 <= offset < 1<<31:
                    error('Label offset must fit in signed 32 bit int. (offset is {})'.format(offset), tok)
                self.setph32(pos, offset % (1<<32))
            else:
                assert 0    # pragma: no cover
        self.lblphs = []
        return self

    def fillRefs(self, pool, error):
        for pos, ref in self.refu8phs:
            self.b[pos] = ref.resolve(pool, error)
        for pos, ref in self.refphs:
            self.b[pos:pos+2] = struct.pack('>H', ref.resolve(pool, error))
        self.refu8phs = []
        self.refphs = []

    def toBytes(self):
        assert not self.refphs and not self.refu8phs
        assert not self._ph8s and not self._ph16s and not self._ph32s
        return bytes(self.b)

    def __len__(self): return len(self.b)

    def __iadd__(self, other):
        # Make sure there are no manual placeholders in other
        assert len(other.lblphs) == len(other._ph8s) + len(other._ph16s) + len(other._ph32s)

        offset = self.pos
        self.b += other.b
        self.refphs.extend((pos + offset, ref) for pos, ref in other.refphs)
        self.refu8phs.extend((pos + offset, ref) for pos, ref in other.refu8phs)
        self.lblphs.extend((pos + offset, lbl, base, dtype) for pos, lbl, base, dtype in other.lblphs)
        self._ph8s.update(pos + offset for pos in other._ph8s)
        self._ph16s.update(pos + offset for pos in other._ph16s)
        self._ph32s.update(pos + offset for pos in other._ph32s)
        return self



#!/usr/bin/env python2
from __future__ import print_function

import os.path, time

import Krakatau
from Krakatau.assembler import parse
from Krakatau import script_util

def assembleSource(source, basename, fatal=False):
    source = source.replace('\t', '  ') + '\n'
    return list(parse.assemble(source, basename, fatal=fatal))

def assembleClass(filename):
    basename = os.path.basename(filename)
    with open(filename, 'rU') as f:
        source = f.read()
    return assembleSource(source, basename)

if __name__== "__main__":
    import argparse
    parser = argparse.ArgumentParser(description='Krakatau bytecode assembler')
    parser.add_argument('-out', help='Path to generate files in')
    parser.add_argument('-r', action='store_true', help="Process all files in the directory target and subdirectories")
    parser.add_argument('-q', action='store_true', help="Only display warnings and errors")
    parser.add_argument('target', help='Name of file to assemble')
    args = parser.parse_args()

    log = script_util.Logger('warning' if args.q else 'info')
    log.info(script_util.copyright)

    out = script_util.makeWriter(args.out, '.class')
    targets = script_util.findFiles(args.target, args.r, '.j')

    start_time = time.time()
    with out:
        for i, target in enumerate(targets):
            log.info('Processing file {}, {}/{} remaining'.format(target, len(targets)-i, len(targets)))

            pairs = assembleClass(target)
            for name, data in pairs:
                filename = out.write(name, data)
                log.info('Class written to', filename)
    print('Total time', time.time() - start_time)



class SSAFunctionBase(object):
    def __init__(self, parent, arguments):
        self.parent = parent
        self.params = list(arguments)
        assert None not in self.params

    def replaceVars(self, rdict):
        self.params = [rdict.get(x,x) for x in self.params]
        assert None not in self.params



reserved_identifiers = '''
abstract
assert
boolean
break
byte
case
catch
char
class
const
continue
default
do
double
else
enum
extends
false
final
finally
float
for
goto
if
implements
import
instanceof
int
interface
long
native
new
null
package
private
protected
public
return
short
static
strictfp
super
switch
synchronized
this
throw
throws
transient
true
try
void
volatile
while
'''.split()



from .. import excepttypes, objtypes
from ..constraints import ObjectConstraint, returnOrThrow, throw
from ..ssa_types import SSA_OBJECT

from .base import BaseOp

class New(BaseOp):
    has_side_effects = True

    def __init__(self, parent, name, inode_key):
        super(New, self).__init__(parent, [], makeException=True)
        self.env = parent.env
        self.tt = objtypes.TypeTT(name, 0)
        self.rval = parent.makeVariable(SSA_OBJECT, origin=self)
        self.rval.uninit_orig_num = inode_key

    def propagateConstraints(self):
        eout = ObjectConstraint.fromTops(self.env, [], (excepttypes.OOM,), nonnull=True)
        rout = ObjectConstraint.fromTops(self.env, [], [self.tt], nonnull=True)
        return returnOrThrow(rout, eout)

class NewArray(BaseOp):
    has_side_effects = True

    def __init__(self, parent, param, baset):
        super(NewArray, self).__init__(parent, [param], makeException=True)
        self.baset = baset
        self.rval = parent.makeVariable(SSA_OBJECT, origin=self)
        self.tt = objtypes.withDimInc(baset, 1)
        self.env = parent.env

    def propagateConstraints(self, i):
        if i.max < 0:
            eout = ObjectConstraint.fromTops(self.env, [], (excepttypes.NegArrSize,), nonnull=True)
            return throw(eout)

        etypes = (excepttypes.OOM,)
        if i.min < 0:
            etypes += (excepttypes.NegArrSize,)

        eout = ObjectConstraint.fromTops(self.env, [], etypes, nonnull=True)
        rout = ObjectConstraint.fromTops(self.env, [], [self.tt], nonnull=True)
        return returnOrThrow(rout, eout)

class MultiNewArray(BaseOp):
    has_side_effects = True

    def __init__(self, parent, params, type_):
        super(MultiNewArray, self).__init__(parent, params, makeException=True)
        self.tt = type_
        self.rval = parent.makeVariable(SSA_OBJECT, origin=self)
        self.env = parent.env

    def propagateConstraints(self, *dims):
        for i in dims:
            if i.max < 0: # ignore possibility of OOM here
                eout = ObjectConstraint.fromTops(self.env, [], (excepttypes.NegArrSize,), nonnull=True)
                return throw(eout)

        etypes = (excepttypes.OOM,)
        for i in dims:
            if i.min < 0:
                etypes += (excepttypes.NegArrSize,)
                break

        eout = ObjectConstraint.fromTops(self.env, [], etypes, nonnull=True)
        rout = ObjectConstraint.fromTops(self.env, [], [self.tt], nonnull=True)
        return returnOrThrow(rout, eout)



from __future__ import print_function

import collections
import math
import re

from ..classfileformat import classdata, mutf8
from ..classfileformat.reader import Reader, TruncatedStreamError
from ..util.thunk import thunk

from . import codes, token_regexes
from . import flags
from .instructions import OPNAMES, OP_CLS, OP_FMIM, OP_LBL, OP_NONE, OP_SHORT

MAX_INLINE_SIZE = 300
MAX_INDENT = 20
WORD_REGEX = re.compile(token_regexes.WORD + r'\Z')

PREFIXES = {'Utf8': 'u', 'Class': 'c', 'String': 's', 'Field': 'f', 'Method': 'm', 'InterfaceMethod': 'im', 'NameAndType': 'nat', 'MethodHandle': 'mh', 'MethodType': 'mt', 'InvokeDynamic': 'id'}

class DisassemblyError(Exception):
    pass

def reprbytes(b):
    # repr will have b in python3 but not python2
    return 'b' + repr(b).lstrip('b')

def isword(s):
    try:
        s = s.decode('ascii')
    except UnicodeDecodeError:
        return False
    return WORD_REGEX.match(s) and s not in flags.FLAGS

def format_string(s):
    try:
        u = mutf8.decode(s)
    except UnicodeDecodeError:
        print('Warning, invalid utf8 data!')
    else:
        if mutf8.encode(u) == s:
            return repr(u).lstrip('u')
    return reprbytes(s)

def make_signed(x, bits):
    if x >= (1 << (bits - 1)):
        x -= 1 << bits
    return x

class StackMapReader(object):
    def __init__(self):
        self.stream = None
        self.tag = -1
        self.pos = -1
        self.count = 0
        self.valid = True

    def setdata(self, r):
        if self.stream is None:
            self.stream = r
            self.count = self.u16() + 1
            self.parseNextPos()
        else:
            # Multiple StackMapTable attributes in same Code attribute
            self.valid = False

    def parseNextPos(self):
        self.count -= 1
        if self.count > 0:
            self.tag = tag = self.u8()

            if tag <= 127: # same and stack_1
                delta = tag % 64
            else: # everything else has 16bit delta field
                delta = self.u16()
            self.pos += delta + 1

    def u8(self):
        try:
            return self.stream.u8()
        except TruncatedStreamError:
            self.valid = False
            return 0

    def u16(self):
        try:
            return self.stream.u16()
        except TruncatedStreamError:
            self.valid = False
            return 0


class ReferencePrinter(object):
    def __init__(self, clsdata, roundtrip):
        self.roundtrip = roundtrip

        self.cpslots = clsdata.pool.slots
        for attr in clsdata.getattrs(b'BootstrapMethods'):
            self.bsslots = classdata.BootstrapMethodsData(attr.stream()).slots
            break
        else:
            self.bsslots = []

        # CP index 0 should always be a raw reference. Additionally, there is one case where exact
        # references are significant due to a bug in the JVM. In the InnerClasses attribute,
        # specifying the same index for inner and outer class will fail verification, but specifying
        # different indexes which point to identical class entries will pass. In this case, we force
        # references to those indexes to be raw, so they don't get merged and break the class.
        self.forcedraw = set()
        for attr in clsdata.getattrs(b'InnerClasses'):
            r = attr.stream()
            for _ in range(r.u16()):
                inner, outer, _, _ = r.u16(), r.u16(), r.u16(), r.u16()
                if inner != outer and clsdata.pool.getclsutf(inner) == clsdata.pool.getclsutf(outer):
                    self.forcedraw.add(inner)
                    self.forcedraw.add(outer)
        self.explicit_forcedraw = self.forcedraw.copy()

        # For invalid cp indices, just output raw ref instead of throwing (including 0)
        for i, slot in enumerate(self.cpslots):
            if slot.tag is None:
                self.forcedraw.add(i)
        self.forcedraw.update(range(len(self.cpslots), 65536))

        self.used = set()
        self.encoded = {}
        self.utfcounts = {}

    def _float_or_double(self, x, nmbits, nebits, suffix, nanfmt):
        nbits = nmbits + nebits + 1
        assert nbits % 32 == 0

        sbit, ebits, mbits = x >> (nbits - 1), (x >> nmbits) % (1 << nebits), x % (1 << nmbits)
        if ebits == (1 << nebits) - 1:
            result = 'NaN' if mbits else 'Infinity'
            if self.roundtrip and mbits:
                result += nanfmt.format(x)
        elif ebits == 0 and mbits == 0:
            result = '0.0'
        else:
            ebias = (1 << (nebits - 1)) - 1
            exponent = ebits - ebias - nmbits
            mantissa = mbits
            if ebits > 0:
                mantissa += 1 << nmbits
            else:
                exponent += 1

            if self.roundtrip:
                result = '0x{:X}p{}'.format(mantissa, exponent)
            else:
                result = repr(math.ldexp(mantissa, exponent))
        return '+-'[sbit] + result + suffix

    def _int(self, x): return str(make_signed(x, 32))
    def _long(self, x): return str(make_signed(x, 64)) + 'L'
    def _float(self, x): return self._float_or_double(x, 23, 8, 'f', '<0x{:08X}>')
    def _double(self, x): return self._float_or_double(x, 52, 11, '', '<0x{:016X}>')

    def _encode_utf(self, ind, wordok=True):
        try:
            return self.encoded[ind][wordok]
        except KeyError:
            s = self.cpslots[ind].data
            string = format_string(s)
            word = s.decode() if isword(s) else string
            self.encoded[ind] = [string, word]
            return word if wordok else string

    def rawref(self, ind, isbs=False):
        return '[{}{}]'.format('bs:' if isbs else '', ind)

    def symref(self, ind, isbs=False):
        self.used.add((ind, isbs))
        if isbs:
            return '[bs:_{}]'.format(ind)

        prefix = PREFIXES.get(self.cpslots[ind].tag, '_')
        return '[{}{}]'.format(prefix, ind)

    def ref(self, ind, isbs=False):
        if self.roundtrip or not isbs and ind in self.forcedraw:
            return self.rawref(ind, isbs)
        return self.symref(ind, isbs)

    def _ident(self, ind, wordok=True):
        if self.cpslots[ind].tag == 'Utf8':
            val = self._encode_utf(ind, wordok=wordok)
            if len(val) < MAX_INLINE_SIZE:
                if len(val) < 50 or self.utfcounts.get(ind, 0) < 10:
                    self.utfcounts[ind] = 1 + self.utfcounts.get(ind, 0)
                    return val

    def utfref(self, ind):
        if self.roundtrip or ind in self.forcedraw:
            return self.rawref(ind)
        temp = self._ident(ind)
        if temp is not None:
            return temp
        return self.symref(ind)

    def clsref(self, ind, tag='Class'):
        assert tag in 'Class Module Package'.split()
        if self.roundtrip or ind in self.forcedraw:
            return self.rawref(ind)
        if self.cpslots[ind].tag == tag:
            ind2 = self.cpslots[ind].refs[0]
            temp = self._ident(ind2)
            if temp is not None:
                return temp
        return self.symref(ind)

    def natref(self, ind):
        if self.roundtrip or ind in self.forcedraw:
            return self.rawref(ind)
        if self.cpslots[ind].tag == 'NameAndType':
            ind2, ind3 = self.cpslots[ind].refs
            temp = self._ident(ind2)
            if temp is not None:
                return temp + ' ' + self.utfref(ind3)
        return self.symref(ind)

    def fmimref(self, ind):
        if self.roundtrip or ind in self.forcedraw:
            return self.rawref(ind)
        if self.cpslots[ind].tag in ['Field', 'Method', 'InterfaceMethod']:
            ind2, ind3 = self.cpslots[ind].refs
            return ' '.join([self.cpslots[ind].tag, self.clsref(ind2), self.natref(ind3)])
        return self.symref(ind)

    def mhnotref(self, ind):
        slot = self.cpslots[ind]
        return codes.handle_rcodes[slot.data] + ' ' + self.taggedref(slot.refs[0], allowed=['Field', 'Method', 'InterfaceMethod'])

    def taggedconst(self, ind):
        slot = self.cpslots[ind]
        if slot.tag == 'Utf8':
            parts = [self._encode_utf(ind)]
        elif slot.tag == 'Int':
            parts = [self._int(slot.data)]
        elif slot.tag == 'Float':
            parts = [self._float(slot.data)]
        elif slot.tag == 'Long':
            parts = [self._long(slot.data)]
        elif slot.tag == 'Double':
            parts = [self._double(slot.data)]
        elif slot.tag in ['Class', 'String', 'MethodType', 'Module', 'Package']:
            parts = [self.utfref(slot.refs[0])]
        elif slot.tag in ['Field', 'Method', 'InterfaceMethod']:
            parts = [self.clsref(slot.refs[0]), self.natref(slot.refs[1])]
        elif slot.tag == 'NameAndType':
            parts = [self.utfref(slot.refs[0]), self.utfref(slot.refs[1])]
        elif slot.tag == 'MethodHandle':
            parts = [self.mhnotref(ind)]
        elif slot.tag == 'InvokeDynamic':
            parts = [self.bsref(slot.refs[0]), self.natref(slot.refs[1])]
        parts.insert(0, slot.tag)
        return ' '.join(parts)

    def taggedref(self, ind, allowed=None):
        if self.roundtrip or ind in self.forcedraw:
            return self.rawref(ind)

        if allowed is None or self.cpslots[ind].tag in allowed:
            temp = self.taggedconst(ind)
            if len(temp) < MAX_INLINE_SIZE:
                return temp
        return self.symref(ind)

    def ldcrhs(self, ind):
        if self.roundtrip or ind in self.forcedraw:
            return self.rawref(ind)
        slot = self.cpslots[ind]
        t = slot.tag

        if t == 'Int':
            return self._int(slot.data)
        elif slot.tag == 'Float':
            return self._float(slot.data)
        elif slot.tag == 'Long':
            return self._long(slot.data)
        elif slot.tag == 'Double':
            return self._double(slot.data)
        elif t == 'String':
            ind2 = self.cpslots[ind].refs[0]
            temp = self._ident(ind2, wordok=False)
            if temp is not None:
                return temp
            return self.symref(ind)
        return self.taggedref(ind, allowed=['Class', 'MethodHandle', 'MethodType'])

    def bsnotref(self, ind, tagged=False):
        slot = self.bsslots[ind]
        parts = []
        if tagged:
            parts.append('Bootstrap')

        if tagged and self.roundtrip:
            parts.append(self.rawref(slot.refs[0]))
        else:
            parts.append(self.mhnotref(slot.refs[0]))
        for bsarg in slot.refs[1:]:
            parts.append(self.taggedref(bsarg))
        parts.append(':')
        return ' '.join(parts)

    def bsref(self, ind):
        if self.roundtrip:
            return self.rawref(ind, isbs=True)
        return self.bsnotref(ind)

LabelInfos = collections.namedtuple('LabelInfos', 'defined used')
class Disassembler(object):
    def __init__(self, clsdata, out, roundtrip):
        self.roundtrip = roundtrip

        self.out = out
        self.cls = clsdata
        self.pool = clsdata.pool

        self.indentlevel = 0
        self.labels = None
        self.refprinter = ReferencePrinter(clsdata, roundtrip)

    def _getattr(a, obj, name):
        for attr in obj.attributes:
            if a.pool.getutf(attr.name) == name:
                return attr

    def sol(a, text=''):
        level = min(a.indentlevel, MAX_INDENT) * 4
        text += ' ' * (level - len(text))
        a.out(text)

    def eol(a): a.out('\n')
    def val(a, s): a.out(s + ' ')
    def int(a, x): a.val(str(x))

    def lbl(a, x):
        a.labels.used.add(x)
        a.val('L{}'.format(x))

    def try_lbl(a, x):
        if a.labels is None or x not in a.labels.defined:
            raise DisassemblyError()
        a.lbl(x)
    ###########################################################################
    def extrablankline(a): a.eol()

    def ref(a, ind, isbs=False): a.val(a.refprinter.ref(ind, isbs))
    def utfref(a, ind): a.val(a.refprinter.utfref(ind))
    def clsref(a, ind, tag='Class'): a.val(a.refprinter.clsref(ind, tag))
    def natref(a, ind): a.val(a.refprinter.natref(ind))
    def fmimref(a, ind): a.val(a.refprinter.fmimref(ind))
    def taggedbs(a, ind): a.val(a.refprinter.bsnotref(ind, tagged=True))
    def taggedconst(a, ind): a.val(a.refprinter.taggedconst(ind))
    def taggedref(a, ind): a.val(a.refprinter.taggedref(ind))
    def ldcrhs(a, ind): a.val(a.refprinter.ldcrhs(ind))

    def flags(a, access, names):
        for i in range(16):
            if access & (1 << i):
                a.val(names[1 << i])

    ###########################################################################
    ### Top level stuff (class, const defs, fields, methods) ##################
    def disassemble(a):
        cls = a.cls
        a.val('.version'), a.int(cls.version[0]), a.int(cls.version[1]), a.eol()
        a.val('.class'), a.flags(cls.access, flags.RFLAGS_CLASS), a.clsref(cls.this), a.eol()
        a.val('.super'), a.clsref(cls.super), a.eol()
        for ref in cls.interfaces:
            a.val('.implements'), a.clsref(ref), a.eol()

        for f in cls.fields:
            a.field(f)

        for m in cls.methods:
            a.method(m)

        for attr in cls.attributes:
            a.attribute(attr)

        a.constdefs()
        a.val('.end class'), a.eol()

    def field(a, f):
        a.val('.field'), a.flags(f.access, flags.RFLAGS_FIELD), a.utfref(f.name), a.utfref(f.desc)

        attrs = f.attributes[:]
        cvattr = a._getattr(f, b'ConstantValue')
        if cvattr and not a.roundtrip:
            a.val('='), a.ldcrhs(cvattr.stream().u16())
            attrs.remove(cvattr)

        if attrs:
            a.val('.fieldattributes'), a.eol()
            a.indentlevel += 1
            for attr in attrs:
                a.attribute(attr)
            a.indentlevel -= 1
            a.val('.end fieldattributes')
        a.eol()

    def method(a, m):
        a.extrablankline()
        a.val('.method'), a.flags(m.access, flags.RFLAGS_METHOD), a.utfref(m.name), a.val(':'), a.utfref(m.desc), a.eol()
        a.indentlevel += 1
        for attr in m.attributes:
            a.attribute(attr, in_method=True)
        a.indentlevel -= 1
        a.val('.end method'), a.eol()

    def constdefs(a):
        if a.roundtrip:
            for ind in range(len(a.refprinter.cpslots)):
                a.constdef(ind, False)
            for ind in range(len(a.refprinter.bsslots)):
                a.constdef(ind, True)
        else:
            assert not a.refprinter.used & a.refprinter.forcedraw
            for ind in sorted(a.refprinter.explicit_forcedraw):
                a.constdef(ind, False)

            done = set()
            while len(done) < len(a.refprinter.used):
                for ind, isbs in sorted(a.refprinter.used - done):
                    a.constdef(ind, isbs)
                    done.add((ind, isbs))

    def constdef(a, ind, isbs):
        if not isbs and a.refprinter.cpslots[ind].tag is None:
            return

        a.sol(), a.val('.bootstrap' if isbs else '.const'), a.ref(ind, isbs), a.val('=')
        if isbs:
            a.taggedbs(ind)
        else:
            a.taggedconst(ind)
        a.eol()

    ###########################################################################
    ### Bytecode ##############################################################
    def code(a, r):
        c = classdata.CodeData(r, a.pool, a.cls.version < (45, 3))
        a.val('.code'), a.val('stack'), a.int(c.stack), a.val('locals'), a.int(c.locals), a.eol()
        a.indentlevel += 1
        assert a.labels is None
        a.labels = LabelInfos(set(), set())

        stackreader = StackMapReader()
        for attr in c.attributes:
            if a.pool.getutf(attr.name) == b'StackMapTable':
                stackreader.setdata(attr.stream())

        rexcepts = c.exceptions[::-1]
        bcreader = Reader(c.bytecode)
        while bcreader.size():
            a.insline_start(bcreader.off, rexcepts, stackreader)
            a.instruction(bcreader)
        a.insline_start(bcreader.off, rexcepts, stackreader), a.eol()

        badlbls = a.labels.used - a.labels.defined
        if badlbls:
            stackreader.valid = False
            a.sol('; Labels used by invalid StackMapTable attribute'), a.eol()
            for pos in sorted(badlbls):
                a.sol('L{}'.format(pos) + ':'), a.eol()

        if stackreader.stream and (stackreader.stream.size() or stackreader.count > 0):
            stackreader.valid = False

        if not stackreader.valid:
            a.sol('.noimplicitstackmap'), a.eol()

        for attr in c.attributes:
            a.attribute(attr, use_raw_stackmap=not stackreader.valid)

        a.labels = None
        a.indentlevel -= 1
        a.sol(), a.val('.end code')

    def insline_start(a, pos, rexcepts, stackreader):
        while rexcepts and rexcepts[-1].start <= pos:
            e = rexcepts.pop()
            a.sol(), a.val('.catch'), a.clsref(e.type), a.val('from'), a.lbl(e.start)
            a.val('to'), a.lbl(e.end), a.val('using'), a.lbl(e.handler), a.eol()

        if stackreader.count > 0 and stackreader.pos == pos:
            r = stackreader
            tag = stackreader.tag
            a.extrablankline()
            a.sol(), a.val('.stack')
            if tag <= 63:
                a.val('same')
            elif tag <= 127:
                a.val('stack_1'), a.verification_type(r)
            elif tag == 247:
                a.val('stack_1_extended'), a.verification_type(r)
            elif tag < 251:
                a.val('chop'), a.int(251 - tag)
            elif tag == 251:
                a.val('same_extended')
            elif tag < 255:
                a.val('append')
                for _ in range(tag - 251):
                    a.verification_type(r)
            else:
                a.val('full')
                a.indentlevel += 1

                a.eol(), a.sol(), a.val('locals')
                for _ in range(r.u16()):
                    a.verification_type(r)
                a.eol(), a.sol(), a.val('stack')
                for _ in range(r.u16()):
                    a.verification_type(r)

                a.indentlevel -= 1
                a.eol(), a.sol(), a.val('.end stack')
            a.eol()
            stackreader.parseNextPos()

        a.sol('L{}'.format(pos) + ':')
        a.labels.defined.add(pos)

    def verification_type(a, r):
        try:
            tag = codes.vt_rcodes[r.u8()]
        except IndexError:
            r.valid = False
            a.val('Top')
            return

        a.val(tag)
        if tag == 'Object':
            a.clsref(r.u16())
        elif tag == 'Uninitialized':
            a.lbl(r.u16())

    def instruction(a, r):
        pos = r.off
        op = OPNAMES[r.u8()]
        a.val(op)

        if op in OP_LBL:
            a.lbl(pos + (r.s32() if op.endswith('_w') else r.s16()))
        elif op in OP_SHORT:
            a.int(r.u8())
        elif op in OP_CLS:
            a.clsref(r.u16())
        elif op in OP_FMIM:
            a.fmimref(r.u16())
        elif op == 'invokeinterface':
            a.fmimref(r.u16()), a.int(r.u8()), r.u8()
        elif op == 'invokedynamic':
            a.taggedref(r.u16()), r.u16()
        elif op in ['ldc', 'ldc_w', 'ldc2_w']:
            a.ldcrhs(r.u8() if op == 'ldc' else r.u16())
        elif op == 'multianewarray':
            a.clsref(r.u16()), a.int(r.u8())
        elif op == 'bipush':
            a.int(r.s8())
        elif op == 'sipush':
            a.int(r.s16())
        elif op == 'iinc':
            a.int(r.u8()), a.int(r.s8())
        elif op == 'wide':
            op2 = OPNAMES[r.u8()]
            a.val(op2), a.int(r.u16())
            if op2 == 'iinc':
                a.int(r.s16())
        elif op == 'newarray':
            a.val(codes.newarr_rcodes[r.u8()])
        elif op == 'tableswitch':
            r.getRaw((3-pos) % 4)
            default = pos + r.s32()
            low, high = r.s32(), r.s32()

            a.int(low), a.eol()
            a.indentlevel += 1
            for _ in range(high - low + 1):
                a.sol(), a.lbl(pos + r.s32()), a.eol()
            a.sol(), a.val('default'), a.val(':'), a.lbl(default), a.eol()
            a.indentlevel -= 1
        elif op == 'lookupswitch':
            r.getRaw((3-pos) % 4)
            default = pos + r.s32()

            a.eol()
            a.indentlevel += 1
            for _ in range(r.s32()):
                a.sol(), a.int(r.s32()), a.val(':'), a.lbl(pos + r.s32()), a.eol()
            a.sol(), a.val('default'), a.val(':'), a.lbl(default), a.eol()
            a.indentlevel -= 1
        else:
            assert op in OP_NONE
        a.eol()

    ###########################################################################
    ### Attributes ############################################################
    def attribute(a, attr, in_method=False, use_raw_stackmap=False):
        name = a.pool.getutf(attr.name)
        if not a.roundtrip and name in (b'BootstrapMethods', b'StackMapTable'):
            return

        # a.extrablankline()
        a.sol()
        isnamed = False
        if a.roundtrip or name is None:
            isnamed = True
            a.val('.attribute'), a.utfref(attr.name)
            if attr.wronglength:
                a.val('length'), a.int(attr.length)

        if name == b'Code' and in_method:
            a.code(attr.stream())
        elif name == b'BootstrapMethods' and a.cls.version >= (51, 0):
            a.val('.bootstrapmethods')
        elif name == b'StackMapTable' and not use_raw_stackmap:
            a.val('.stackmaptable')
        elif a.attribute_fallible(name, attr):
            pass
        else:
            print('Nonstandard attribute', name[:70], len(attr.raw))
            if not isnamed:
                a.val('.attribute'), a.utfref(attr.name)
            a.val(reprbytes(attr.raw))

        a.eol()

    def attribute_fallible(a, name, attr):
        # Temporarily buffer output so we don't get partial output if attribute disassembly fails
        # in case of failure, we'll fall back to binary output in the caller
        orig_out = a.out
        buffer_ = []
        a.out = buffer_.append

        try:
            r = attr.stream()
            if name == b'AnnotationDefault':
                a.val('.annotationdefault'), a.element_value(r)
            elif name == b'ConstantValue':
                a.val('.constantvalue'), a.ldcrhs(r.u16())
            elif name == b'Deprecated':
                a.val('.deprecated')
            elif name == b'EnclosingMethod':
                a.val('.enclosing method'), a.clsref(r.u16()), a.natref(r.u16())
            elif name == b'Exceptions':
                a.val('.exceptions')
                for _ in range(r.u16()):
                    a.clsref(r.u16())
            elif name == b'InnerClasses':
                a.indented_line_list(r, a._innerclasses_item, 'innerclasses')
            elif name == b'LineNumberTable':
                a.indented_line_list(r, a._linenumber_item, 'linenumbertable')
            elif name == b'LocalVariableTable':
                a.indented_line_list(r, a._localvariabletable_item, 'localvariabletable')
            elif name == b'LocalVariableTypeTable':
                a.indented_line_list(r, a._localvariabletable_item, 'localvariabletypetable')
            elif name == b'MethodParameters':
                a.indented_line_list(r, a._methodparams_item, 'methodparameters', bytelen=True)
            elif name == b'Module':
                a.module_attr(r)
            elif name == b'ModuleMainClass':
                a.val('.modulemainclass'), a.clsref(r.u16())
            elif name == b'ModulePackages':
                a.val('.modulepackages')
                for _ in range(r.u16()):
                    a.clsref(r.u16(), tag='Package')
            elif name in (b'RuntimeVisibleAnnotations', b'RuntimeVisibleParameterAnnotations',
                b'RuntimeVisibleTypeAnnotations', b'RuntimeInvisibleAnnotations',
                b'RuntimeInvisibleParameterAnnotations', b'RuntimeInvisibleTypeAnnotations'):
                a.val('.runtime')
                a.val('invisible' if b'Inv' in name else 'visible')
                if b'Type' in name:
                    a.val('typeannotations'), a.eol()
                    a.indented_line_list(r, a.type_annotation_line, 'runtime', False)
                elif b'Parameter' in name:
                    a.val('paramannotations'), a.eol()
                    a.indented_line_list(r, a.param_annotation_line, 'runtime', False, bytelen=True)
                else:
                    a.val('annotations'), a.eol()
                    a.indented_line_list(r, a.annotation_line, 'runtime', False)

            elif name == b'Signature':
                a.val('.signature'), a.utfref(r.u16())
            elif name == b'SourceDebugExtension':
                a.val('.sourcedebugextension')
                a.val(reprbytes(attr.raw))
            elif name == b'SourceFile':
                a.val('.sourcefile'), a.utfref(r.u16())
            elif name == b'Synthetic':
                a.val('.synthetic')

            # check for extra data in the attribute
            if r.size():
                raise DisassemblyError()

        except (TruncatedStreamError, DisassemblyError):
            a.out = orig_out
            return False

        a.out = orig_out
        a.out(''.join(buffer_))
        return True

    def module_attr(a, r):
        a.val('.module'), a.clsref(r.u16(), tag='Module'), a.flags(r.u16(), flags.RFLAGS_MOD_OTHER)
        a.val('version'), a.utfref(r.u16()), a.eol()
        a.indentlevel += 1

        for _ in range(r.u16()):
            a.sol(), a.val('.requires'), a.clsref(r.u16(), tag='Module'), a.flags(r.u16(), flags.RFLAGS_MOD_REQUIRES), a.val('version'), a.utfref(r.u16()), a.eol()

        for dir_ in ('.exports', '.opens'):
            for _ in range(r.u16()):
                a.sol(), a.val(dir_), a.clsref(r.u16(), tag='Package'), a.flags(r.u16(), flags.RFLAGS_MOD_OTHER)
                count = r.u16()
                if count:
                    a.val('to')
                    for _ in range(count):
                        a.clsref(r.u16(), tag='Module')
                a.eol()

        for _ in range(r.u16()):
            a.sol(), a.val('.uses'), a.clsref(r.u16()), a.eol()

        for _ in range(r.u16()):
            a.sol(), a.val('.provides'), a.clsref(r.u16()), a.val('with')
            for _ in range(r.u16()):
                a.clsref(r.u16())
            a.eol()
        a.indentlevel -= 1
        a.sol(), a.val('.end module')

    def indented_line_list(a, r, cb, dirname, dostart=True, bytelen=False):
        if dostart:
            a.val('.' + dirname), a.eol()
        a.indentlevel += 1
        for _ in range(r.u8() if bytelen else r.u16()):
            a.sol(), cb(r), a.eol()
        a.indentlevel -= 1
        if dirname is not None:
            a.sol(), a.val('.end ' + dirname)

    def _innerclasses_item(a, r): a.clsref(r.u16()), a.clsref(r.u16()), a.utfref(r.u16()), a.flags(r.u16(), flags.RFLAGS_CLASS)
    def _linenumber_item(a, r): a.try_lbl(r.u16()), a.int(r.u16())
    def _localvariabletable_item(a, r):
        start, length, name, desc, ind = r.u16(), r.u16(), r.u16(), r.u16(), r.u16()
        a.int(ind), a.val('is'), a.utfref(name), a.utfref(desc),
        a.val('from'), a.try_lbl(start), a.val('to'), a.try_lbl(start + length)
    def _methodparams_item(a, r): a.utfref(r.u16()), a.flags(r.u16(), flags.RFLAGS_MOD_OTHER)

    ###########################################################################
    ### Annotations ###########################################################
    def annotation_line(a, r):
        a.val('.annotation'), a.annotation_contents(r), a.sol(), a.val('.end'), a.val('annotation')

    def param_annotation_line(a, r):
        a.indented_line_list(r, a.annotation_line, 'paramannotation')

    def type_annotation_line(a, r):
        a.val('.typeannotation')
        a.indentlevel += 1
        a.ta_target_info(r) # Note: begins on same line as .typeannotation
        a.ta_target_path(r)
        a.sol(), a.annotation_contents(r),
        a.indentlevel -= 1
        a.sol(), a.val('.end'), a.val('typeannotation')

    def ta_target_info(a, r):
        tag = r.u8()
        a.int(tag)
        if tag <= 0x01:
            a.val('typeparam'), a.int(r.u8())
        elif tag <= 0x10:
            a.val('super'), a.int(r.u16())
        elif tag <= 0x12:
            a.val('typeparambound'), a.int(r.u8()), a.int(r.u8())
        elif tag <= 0x15:
            a.val('empty')
        elif tag <= 0x16:
            a.val('methodparam'), a.int(r.u8())
        elif tag <= 0x17:
            a.val('throws'), a.int(r.u16())
        elif tag <= 0x41:
            a.val('localvar'), a.eol()
            a.indented_line_list(r, a._localvarrange, 'localvar', False)
        elif tag <= 0x42:
            a.val('catch'), a.int(r.u16())
        elif tag <= 0x46:
            a.val('offset'), a.try_lbl(r.u16())
        else:
            a.val('typearg'), a.try_lbl(r.u16()), a.int(r.u8())
        a.eol()

    def _localvarrange(a, r):
        start, length, index = r.u16(), r.u16(), r.u16()
        if start == length == 0xFFFF: # WTF, Java?
            a.val('nowhere')
        else:
            a.val('from'), a.try_lbl(start), a.val('to'), a.try_lbl(start + length)
        a.int(index)

    def ta_target_path(a, r):
        a.sol(), a.indented_line_list(r, a._type_path_segment, 'typepath', bytelen=True), a.eol()

    def _type_path_segment(a, r):
        a.int(r.u8()), a.int(r.u8())

    # The following are recursive and can be nested arbitrarily deep,
    # so we use generators and a thunk to avoid the Python stack limit.
    def element_value(a, r): thunk(a._element_value(r))
    def annotation_contents(a, r): thunk(a._annotation_contents(r))

    def _element_value(a, r):
        tag = codes.et_rtags.get(r.u8())
        if tag is None:
            raise DisassemblyError()
        a.val(tag)
        if tag == 'annotation':
            (yield a._annotation_contents(r)), a.sol(), a.val('.end'), a.val('annotation')
        elif tag == 'array':
            a.eol()
            a.indentlevel += 1
            for _ in range(r.u16()):
                a.sol(), (yield a._element_value(r)), a.eol()
            a.indentlevel -= 1
            a.sol(), a.val('.end'), a.val('array')
        elif tag == 'enum':
            a.utfref(r.u16()), a.utfref(r.u16())
        elif tag == 'class' or tag == 'string':
            a.utfref(r.u16())
        else:
            a.ldcrhs(r.u16())

    def _annotation_contents(a, r):
        a.utfref(r.u16()), a.eol()
        a.indentlevel += 1
        for _ in range(r.u16()):
            a.sol(), a.utfref(r.u16()), a.val('='), (yield a._element_value(r)), a.eol()
        a.indentlevel -= 1



import re

# First alternative handles a single surrogate, in case input string somehow contains unmerged surrogates
NONASTRAL_REGEX = re.compile(u'[\ud800-\udfff]|[\0-\ud7ff\ue000-\uffff]+')

def encode(s):
    assert not isinstance(s, bytes)
    b = b''
    pos = 0
    while pos < len(s):
        x = ord(s[pos])
        if x >= 1<<16:
            x -= 1<<16
            high = 0xD800 + (x >> 10)
            low = 0xDC00 + (x % (1 << 10))
            b += unichr(high).encode('utf8')
            b += unichr(low).encode('utf8')
            pos += 1
        else:
            m = NONASTRAL_REGEX.match(s, pos)
            b += m.group().encode('utf8')
            pos = m.end()
    return b.replace(b'\0', b'\xc0\x80')

# Warning, decode(encode(s)) != s if s contains astral characters, as they are converted to surrogate pairs
def decode(b):
    assert isinstance(b, bytes)
    return b.replace(b'\xc0\x80', b'\0').decode('utf8')



from ... import floatutil as fu

from ..mixin import ValueType

SPECIALS = frozenset((fu.NAN, fu.INF, fu.NINF, fu.ZERO, fu.NZERO))

def botRange(size):
    mbits, emin, emax = size
    mag = (1<<(mbits+1))-1, emax-mbits
    return (-1,mag), (1,mag)

class FloatConstraint(ValueType):
    def __init__(self, size, finite, special):
        self.size = size
        self.finite = finite
        self.spec = special

        self.isBot = (special == SPECIALS) and (finite == botRange(size))

    @staticmethod
    def const(size, val):
        if val in SPECIALS:
            return FloatConstraint(size, (None, None), frozenset([val]))
        return FloatConstraint(size, (val, val), frozenset())

    @staticmethod
    def bot(size):
        finite = botRange(size)
        return FloatConstraint(size, finite, SPECIALS)

    def _key(self): return self.finite, self.spec

    def join(*cons): # more precise (intersection)
        spec = frozenset.intersection(*[c.spec for c in cons])
        ranges = [c.finite for c in cons]

        if (None, None) in ranges:
            xmin = xmax = None
        else:
            mins, maxs = zip(*ranges)
            xmin = max(mins, key=fu.sortkey)
            xmax = min(maxs, key=fu.sortkey)
            if fu.sortkey(xmax) < fu.sortkey(xmin):
                xmin = xmax = None
        if not xmin and not spec:
            return None
        return FloatConstraint(cons[0].size, (xmin, xmax), spec)

    def meet(*cons):
        spec = frozenset.union(*[c.spec for c in cons])
        ranges = [c.finite for c in cons if c.finite != (None,None)]

        if ranges:
            mins, maxs = zip(*ranges)
            xmin = min(mins, key=fu.sortkey)
            xmax = max(maxs, key=fu.sortkey)
        else:
            xmin = xmax = None
        return FloatConstraint(cons[0].size, (xmin, xmax), spec)



from . import constant_pool, field, method
from .attributes_raw import fixAttributeNames, get_attributes_raw

cp_structFmts = {3: '>i',
                4: '>i',    # floats and doubles internally represented as integers with same bit pattern
                5: '>q',
                6: '>q',
                7: '>H',
                8: '>H',
                9: '>HH',
                10: '>HH',
                11: '>HH',
                12: '>HH',
                15: '>BH',
                16: '>H',
                18: '>HH'}

def get_cp_raw(bytestream):
    const_count = bytestream.get('>H')
    assert const_count > 1

    placeholder = None,None
    pool = [placeholder]

    while len(pool) < const_count:
        tag = bytestream.get('B')
        if tag == 1: # utf8
            length = bytestream.get('>H')
            data = bytestream.getRaw(length)
            val = tag, (data,)
        else:
            val = tag,bytestream.get(cp_structFmts[tag], True)
        pool.append(val)
        # Longs and Doubles take up two spaces in the pool
        if tag == 5 or tag == 6:
            pool.append(placeholder)
    assert len(pool) == const_count
    return pool

def get_field_raw(bytestream):
    flags, name, desc = bytestream.get('>HHH')
    attributes = get_attributes_raw(bytestream)
    return flags, name, desc, attributes

def get_fields_raw(bytestream):
    count = bytestream.get('>H')
    return [get_field_raw(bytestream) for _ in range(count)]

# fields and methods have same raw format
get_method_raw = get_field_raw
get_methods_raw = get_fields_raw

class ClassFile(object):
    flagVals = {'PUBLIC':0x0001,
                'FINAL':0x0010,
                'SUPER':0x0020,
                'INTERFACE':0x0200,
                'ABSTRACT':0x0400,
                'SYNTHETIC':0x1000,
                'ANNOTATION':0x2000,
                'ENUM':0x4000,

                # These flags are only used for InnerClasses attributes
                'PRIVATE':0x0002,
                'PROTECTED':0x0004,
                'STATIC':0x0008,
                }

    def __init__(self, bytestream):
        magic, minor, major = bytestream.get('>LHH')
        assert magic == 0xCAFEBABE
        self.version = major,minor

        const_pool_raw = get_cp_raw(bytestream)
        flags, self.this, self.super = bytestream.get('>HHH')

        interface_count = bytestream.get('>H')
        self.interfaces_raw = [bytestream.get('>H') for _ in range(interface_count)]

        self.fields_raw = get_fields_raw(bytestream)
        self.methods_raw = get_methods_raw(bytestream)

        ic_indices = [i for i,x in enumerate(const_pool_raw) if x == (1, ("InnerClasses",))]
        self.attributes_raw = get_attributes_raw(bytestream, ic_indices)
        assert bytestream.size() == 0

        self.flags = frozenset(name for name,mask in ClassFile.flagVals.items() if (mask & flags))
        self.cpool = constant_pool.ConstPool(const_pool_raw)
        self.name = self.cpool.getArgsCheck('Class', self.this)
        self.elementsLoaded = False

        self.env = self.supername = None
        self.fields = self.methods = self.attributes = None
        if self.super:
            self.supername = self.cpool.getArgsCheck('Class', self.super)

    def loadElements(self, keepRaw=False):
        if self.elementsLoaded:
            return
        self.fields = [field.Field(m, self, keepRaw) for m in self.fields_raw]
        self.methods = [method.Method(m, self, keepRaw) for m in self.methods_raw]
        self.attributes = fixAttributeNames(self.attributes_raw, self.cpool)

        self.fields_raw = self.methods_raw = None
        if not keepRaw:
            self.attributes_raw = None
        self.elementsLoaded = True



import collections
import itertools

from . import objtypes
from .mixin import ValueType

class CatchSetManager(object):
    def __init__(self, env, sets, mask):
        self.env, self.sets, self.mask = env, sets, mask
        assert not self._conscheck()

    @staticmethod # factory
    def new(env, chpairs):
        sets = collections.OrderedDict() # make this ordered since OnException relies on it
        sofar = empty = ExceptionSet.EMPTY
        for catchtype, handler in chpairs:
            old = sets.get(handler, empty)
            new = ExceptionSet.fromTops(env, catchtype)
            sets[handler] = old | (new - sofar)
            sofar = sofar | new
        return CatchSetManager(env, sets, sofar)

    def newMask(self, mask):
        for k in self.sets:
            self.sets[k] &= mask
        self.mask &= mask
        assert not self._conscheck()

    def pruneKeys(self):
        for handler, catchset in list(self.sets.items()):
            if not catchset:
                del self.sets[handler]

    def copy(self):
        return CatchSetManager(self.env, self.sets.copy(), self.mask)

    def replaceKeys(self, replace):
        self.sets = collections.OrderedDict((replace.get(key,key), val) for key, val in self.sets.items())

    def _conscheck(self):
        temp = ExceptionSet.EMPTY
        for v in self.sets.values():
            assert not v & temp
            temp |= v
        assert temp == self.mask
        assert isinstance(self.sets, collections.OrderedDict)

class ExceptionSet(ValueType):
    __slots__ = "env pairs".split()
    def __init__(self, env, pairs): # assumes arguments are in reduced form
        self.env = env
        self.pairs = frozenset([(x,frozenset(y)) for x,y in pairs])

        # We allow env to be None for the empty set so we can construct empty sets easily
        # Any operation resulting in a nonempty set will get its env from the nonempty argument
        assert self.empty() or self.env is not None

        # make sure set is fully reduced
        parts = []
        for t, holes in pairs:
            parts.append(t)
            parts.extend(holes)
        assert len(set(parts)) == len(parts)

    @staticmethod # factory
    def fromTops(env, *tops):
        return ExceptionSet(env, [(x, frozenset()) for x in tops])

    def _key(self): return self.pairs
    def empty(self): return not self.pairs
    def __nonzero__(self): return bool(self.pairs)

    def getTopTTs(self): return sorted([objtypes.TypeTT(top,0) for (top,holes) in self.pairs])

    def __sub__(self, other):
        assert type(self) == type(other)
        if self.empty() or other.empty():
            return self
        if self == other:
            return ExceptionSet.EMPTY

        subtest = self.env.isSubclass
        pairs = self.pairs

        for pair2 in other.pairs:
            # Warning, due to a bug in Python, TypeErrors raised inside the gen expr will give an incorect error message
            # TypeError: type object argument after * must be a sequence, not generator
            # This can be worked around by using a list comprehension instead of a genexpr after the *
            pairs = itertools.chain(*[ExceptionSet.diffPair(subtest, pair1, pair2) for pair1 in pairs])
        return ExceptionSet.reduce(self.env, pairs)

    def __or__(self, other):
        assert type(self) == type(other)
        if other.empty() or self == other:
            return self
        if self.empty():
            return other
        return ExceptionSet.reduce(self.env, self.pairs | other.pairs)

    def __and__(self, other):
        assert type(self) == type(other)
        new = self - (self - other)
        return new

    def isdisjoint(self, other):
        return (self-other) == self

    def __str__(self):   # pragma: no cover
        parts = [('{} - [{}]'.format(top, ', '.join(sorted(holes))) if holes else top) for top, holes in self.pairs]
        return 'ES[{}]'.format(', '.join(parts))
    __repr__ = __str__

    @staticmethod
    def diffPair(subtest, pair1, pair2): # subtract pair2 from pair1. Returns a list of new pairs
        # todo - find way to make this less ugly
        t1, holes1 = pair1
        t2, holes2 = pair2
        if subtest(t1,t2): # t2 >= t1
            if any(subtest(t1, h) for h in holes2):
                return pair1,
            else:
                newpairs = []
                holes2 = [h for h in holes2 if subtest(h, t1) and not any(subtest(h,h2) for h2 in holes1)]

                for h in holes2:
                    newholes = [h2 for h2 in holes1 if subtest(h2, h)]
                    newpairs.append((h, newholes))
                return newpairs
        elif subtest(t2,t1): # t2 < t1
            if any(subtest(t2, h) for h in holes1):
                return pair1,
            else:
                newpairs = [(t1,ExceptionSet.reduceHoles(subtest, list(holes1)+[t2]))]
                holes2 = [h for h in holes2 if not any(subtest(h,h2) for h2 in holes1)]

                for h in holes2:
                    newholes = [h2 for h2 in holes1 if subtest(h2, h)]
                    newpairs.append((h, newholes))
                return newpairs
        else:
            return pair1,

    @staticmethod
    def mergePair(subtest, pair1, pair2): # merge pair2 into pair1 and return the union
        t1, holes1 = pair1
        t2, holes2 = pair2
        assert subtest(t2,t1)

        if t2 in holes1:
            holes1 = list(holes1)
            holes1.remove(t2)
            return t1, holes1 + list(holes2)

        # TODO - this can probably be made more efficient
        holes1a = set(h for h in holes1 if not subtest(h, t2))
        holes1b = [h for h in holes1 if h not in holes1a]

        merged_holes = set()
        for h1, h2 in itertools.product(holes1b, holes2):
            if subtest(h2, h1):
                merged_holes.add(h1)
            elif subtest(h1, h2):
                merged_holes.add(h2)
        merged_holes = ExceptionSet.reduceHoles(subtest, merged_holes)
        assert len(merged_holes) <= len(holes1b) + len(holes2)
        return t1, (list(holes1a) + merged_holes)

    @staticmethod
    def reduceHoles(subtest, holes):
        newholes = []
        for hole in holes:
            for ehole in newholes:
                if subtest(hole, ehole):
                    break
            else:
                newholes = [hole] + [h for h in newholes if not subtest(h, hole)]
        return newholes

    @staticmethod
    def reduce(env, pairs):
        subtest = env.isSubclass
        pairs = [pair for pair in pairs if pair[0] not in pair[1]] # remove all degenerate pairs

        newpairs = []
        while pairs:
            top, holes = pair = pairs.pop()

            # look for an existing top to merge into
            for epair in newpairs[:]:
                etop, eholes = epair
                # new pair can be merged into existing pair
                if subtest(top, etop) and (top in eholes or not any(subtest(top, ehole) for ehole in eholes)):
                    new = ExceptionSet.mergePair(subtest, epair, pair)
                    newpairs, pairs = [new], [p for p in newpairs if p is not epair] + pairs
                    break
                # existing pair can be merged into new pair
                elif subtest(etop, top) and (etop in holes or not any(subtest(etop, hole) for hole in holes)):
                    new = ExceptionSet.mergePair(subtest, pair, epair)
                    newpairs, pairs = [new], [p for p in newpairs if p is not epair] + pairs
                    break
            # pair is incomparable to all existing pairs
            else:
                holes = ExceptionSet.reduceHoles(subtest, holes)
                newpairs.append((top,holes))
        return ExceptionSet(env, newpairs)

ExceptionSet.EMPTY = ExceptionSet(None, [])
